#ifndef lint
static char yysccsid[] = "@(#)yaccpar	1.9 (Berkeley) 02/21/93";
#endif
#define YYBYACC 1
#define YYMAJOR 1
#define YYMINOR 9
#define yyclearin (yychar=(-1))
#define yyerrok (yyerrflag=0)
#define YYRECOVERING (yyerrflag!=0)
#define YYPREFIX "yy"
#line 2 "gram.y"
/* $Header: /local/src/CVS/nickle/gram.y,v 1.137 2004/06/17 20:46:09 keithp Exp $ */

/*
 * Copyright © 1988-2004 Keith Packard and Bart Massey.
 * All Rights Reserved.  See the file COPYING in this directory
 * for licensing information.
 */

#include	"nickle.h"
#include	<stdio.h>

/*
 * This grammar generates 4 reduce/reduce and 4 shift/reduce conflicts:
 * 
 * reduce/reduce conflicts on:
 *
 *	    *int . func 
 *	    **int . func
 *
 *	Is the '*' a dereference operator or a pointer type modifier?
 *	The grammar is ordered to make this a type modifier since
 *	the other way is less common.  Use parens when you want this.
 *
 *  reduce/reduce conflicts on:
 *
 *	   &int . func
 *	   &&int . func
 *
 *	Is the '&' a reference operator or a reference type modifier?
 *	The grammar is ordered to make this a type modifier since
 *	the other way is less common.  Use parens when you want this.
 *
 *  shift/reduce conflict on NL:
 *
 *	    opt_nl : .
 *
 *	This says that you can have multiple NLs in a row
 *	and the grammar can't tell which opt_nl is reduced.
 *	No big deal.  (Could be fixed by replacing with opt_nls
 *      everywhere, but would change the semantics? --Bart)
 *
 *  shift/reduce conflict on CATCH:
 *
 *	That's because CATCH blocks are optional and chained.
 *	Shifting binds the CATCH to the nearest enclosing TRY,
 *	just like ELSE
 *
 *  shift/reduce conflict in ASSIGN in initializers
 *
 *	That's because struct initializers look like assignment
 *	expressions while array initializers permit any expression.
 *	Shift yields struct initialization, so the effect is
 *	to make assignment expressions invalid in array initializers
 *	(of course, you can always enclose an assignment in parens)
 *
 *  shift/reduce conflict on ELSE
 *
 *	Need we say more?
 */
int ignorenl;
int notCommand;
int seeComment;
NamespacePtr	LexNamespace;
int funcDepth;

void ParseError (char *fmt, ...);
void yyerror (char *msg);

typedef enum _CanonTypeResult {
    CanonTypeUndefined,
    CanonTypeForward,
    CanonTypeDefined,
} CanonTypeResult;
    
static Expr *
ParseCanonFor (Expr *expr);

static CanonTypeResult
ParseCanonType (TypePtr type, Bool forwardAllowed);

static SymbolPtr
ParseNewSymbol (Publish publish, Class class, Type *type, Atom name);

#line 87 "gram.y"
typedef union {
    int		    ints;
    Value	    value;
    Class	    class;
    ArgType	    *argType;
    Type	    *type;
    Publish	    publish;
    ExprPtr	    expr;
    Atom	    atom;
    DeclListPtr	    declList;
    MemListPtr	    memList;
    Fulltype	    fulltype;
    ArgDecl	    argDecl;
    SymbolPtr	    symbol;
    NamespacePtr    namespace;
    CodePtr	    code;
    Bool	    bool;
    AtomListPtr	    atomList;
    FuncDecl	    funcDecl;
} YYSTYPE;
#line 117 "gram.c"
#define VAR 257
#define EXPR 258
#define ARRAY 259
#define STRUCT 260
#define UNION 261
#define ENUM 262
#define COMP 263
#define HASH 264
#define NL 265
#define SEMI 266
#define MOD 267
#define OC 268
#define CC 269
#define DOLLAR 270
#define DOTDOTDOT 271
#define ENDFILE 272
#define GLOBAL 273
#define AUTO 274
#define STATIC 275
#define CONST 276
#define POLY 277
#define INTEGER 278
#define NATURAL 279
#define RATIONAL 280
#define REAL 281
#define STRING 282
#define FILET 283
#define MUTEX 284
#define SEMAPHORE 285
#define CONTINUATION 286
#define THREAD 287
#define VOID 288
#define BOOL 289
#define FUNCTION 290
#define FUNC 291
#define EXCEPTION 292
#define RAISE 293
#define TYPEDEF 294
#define IMPORT 295
#define NEW 296
#define ANONINIT 297
#define NAMESPACE 298
#define PUBLIC 299
#define PROTECTED 300
#define EXTEND 301
#define IF 302
#define ELSE 303
#define WHILE 304
#define DO 305
#define FOR 306
#define SWITCH 307
#define BREAK 308
#define CONTINUE 309
#define RETURNTOK 310
#define FORK 311
#define CASE 312
#define DEFAULT 313
#define TRY 314
#define CATCH 315
#define TWIXT 316
#define NAME 317
#define TYPENAME 318
#define NAMESPACENAME 319
#define COMMAND 320
#define NAMECOMMAND 321
#define TEN_NUM 322
#define OCTAL0_NUM 323
#define OCTAL_NUM 324
#define BINARY_NUM 325
#define HEX_NUM 326
#define TEN_FLOAT 327
#define OCTAL0_FLOAT 328
#define OCTAL_FLOAT 329
#define BINARY_FLOAT 330
#define HEX_FLOAT 331
#define CHAR_CONST 332
#define STRING_CONST 333
#define POLY_CONST 334
#define THREAD_CONST 335
#define COMMENT_CONST 336
#define VOIDVAL 337
#define BOOLVAL 338
#define DARROW 339
#define POUND 340
#define COMMA 341
#define ASSIGN 342
#define ASSIGNPLUS 343
#define ASSIGNMINUS 344
#define ASSIGNTIMES 345
#define ASSIGNDIVIDE 346
#define ASSIGNDIV 347
#define ASSIGNMOD 348
#define ASSIGNPOW 349
#define ASSIGNSHIFTL 350
#define ASSIGNSHIFTR 351
#define ASSIGNLXOR 352
#define ASSIGNLAND 353
#define ASSIGNLOR 354
#define QUEST 355
#define COLON 356
#define OR 357
#define AND 358
#define LOR 359
#define LXOR 360
#define LAND 361
#define EQ 362
#define NE 363
#define LT 364
#define GT 365
#define LE 366
#define GE 367
#define SHIFTL 368
#define SHIFTR 369
#define PLUS 370
#define MINUS 371
#define TIMES 372
#define DIVIDE 373
#define DIV 374
#define POW 375
#define UNIONCAST 376
#define UMINUS 377
#define BANG 378
#define FACT 379
#define LNOT 380
#define INC 381
#define DEC 382
#define STAR 383
#define STARSTAR 384
#define AMPER 385
#define THREADID 386
#define OS 387
#define CS 388
#define DOT 389
#define ARROW 390
#define STAROS 391
#define CALL 392
#define OP 393
#define CP 394
#define POINTER 395
#define COLONCOLON 396
#define YYERRCODE 256
short yylhs[] = {                                        -1,
    0,    0,   90,   90,   93,   93,   93,   94,   95,   92,
   96,   97,   97,   98,   98,   99,   99,   91,   91,   91,
   91,   91,   91,   91,    2,    2,    4,    4,    3,    6,
    6,    6,    5,    5,   22,   22,    1,    1,   23,   23,
   54,   54,  100,  101,    7,  102,  103,   11,   11,   12,
   12,   12,   12,   12,   12,   12,   12,   12,   12,   12,
   12,   12,   12,   12,   12,   12,   12,  104,   12,   12,
   12,   12,   64,   64,   13,   13,   14,    8,    8,  105,
    9,   15,   15,  106,   75,   76,   76,   16,   17,   17,
   18,   18,   19,   20,   20,   21,   21,   32,   32,   28,
   28,   30,   30,   35,   35,   29,   29,   31,   38,   33,
   34,   34,   34,   36,   36,   36,   37,   37,   37,   37,
   37,   37,   37,   41,   41,   41,   41,   41,   41,   40,
   40,   40,   40,   40,   42,   42,   42,   42,   42,   42,
   39,   39,   46,   46,   46,   46,   46,   46,   46,   46,
   46,   46,   46,   46,   43,   43,   44,   44,   45,   45,
   47,   47,   48,   48,   49,   49,   50,   50,   50,   50,
   51,   51,   52,   52,   53,   53,   55,   55,   56,   56,
   57,   57,   60,   58,   58,   59,   59,   61,   61,   62,
   62,   63,   63,   65,   65,   71,   71,   66,   66,   67,
   67,   69,   69,   70,   70,   10,  107,   10,   68,   68,
   68,   68,   68,   68,   68,   68,   68,   68,   68,   68,
   68,   68,   68,   68,   68,   68,   68,   68,   68,   68,
   68,   68,   68,   68,   68,   68,   68,   68,   68,   68,
   68,   68,   68,   68,   68,   68,   68,   68,   68,   68,
   68,   68,   68,   68,   68,   68,   68,   68,   68,   68,
   68,   68,   68,   68,   68,   68,   68,   68,   68,   68,
   68,   68,   68,   68,   68,   68,   73,   73,   72,   72,
   72,   72,   72,   72,   72,   72,   72,   72,   72,   72,
   72,   74,   74,   74,   74,   74,   77,   77,   77,   78,
   78,   24,   24,   25,   25,   26,   27,   79,   79,   80,
   80,   81,   81,   81,   82,   83,   83,   84,   84,   85,
   85,   86,   87,   87,   88,   88,   89,   89,   89,   89,
};
short yylen[] = {                                         2,
    2,    0,    1,    3,    1,    1,    1,    0,    0,    0,
    0,    1,    0,    1,    0,    1,    0,    4,    6,    4,
    6,    6,    1,    1,    1,    0,    3,    1,    2,    1,
    1,    1,    3,    0,    2,    1,    2,    1,    3,    2,
    1,    1,    0,    0,    3,    2,    2,    2,    0,    9,
   11,    9,   10,    9,    9,   10,    2,    2,    3,    2,
    1,    1,    4,   10,    6,    6,    7,    0,    9,    6,
    5,   11,    3,    1,    2,    0,    7,    1,    1,    0,
    2,    1,    4,    0,    2,    1,    0,    3,    2,    0,
    4,    3,    3,    2,    0,    7,    3,    1,    0,    3,
    1,    3,    1,    1,    1,    4,    2,    1,    1,    4,
    3,    4,    4,    2,    2,    0,    4,    3,    3,    2,
    3,    2,    2,    2,    4,    4,    4,    4,    0,    2,
    2,    2,    2,    2,    1,    4,    4,    4,    3,    1,
    1,    0,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    0,    3,    1,    3,    1,
    3,    1,    4,    0,    4,    0,    1,    1,    1,    1,
    1,    0,    1,    1,    1,    1,    3,    2,    3,    2,
    2,    1,    3,    2,    1,    3,    2,    2,    1,    1,
    0,    1,    0,    1,    2,    1,    3,    1,    0,    3,
    1,    1,    0,    3,    2,    3,    0,    2,    3,    7,
    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,
    2,    2,    2,    2,    3,    3,    3,    3,    3,    3,
    3,    3,    3,    5,    3,    3,    3,    3,    3,    3,
    3,    3,    3,    3,    3,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    2,    6,    8,    8,    8,    3,    6,    2,    1,    3,
    3,    4,    4,    4,    3,    3,    1,    0,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    2,    0,    5,
    9,    2,    1,    3,    1,    1,    0,    3,    1,    1,
    1,    1,    2,    0,    4,    3,    1,    3,    2,    1,
    1,    4,    3,    1,    3,    3,    1,    1,    1,    2,
};
short yydefred[] = {                                      2,
    0,   10,   23,   24,    1,    3,    0,    0,    0,    0,
    0,   62,    0,   43,    0,  167,  168,  169,  170,  143,
  144,  145,  146,  147,  148,  149,  150,  151,  152,  153,
  154,    8,    0,  173,  174,  176,    8,    8,    8,    8,
    8,    0,    0,    0,    0,    8,    8,   41,   36,    0,
    0,    0,  247,  249,  248,  250,  251,  252,  254,  253,
  255,  256,  257,  258,  259,  260,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  269,    0,  246,   61,
   10,  140,    0,   84,   43,    0,    0,    0,    0,  135,
    0,    0,    0,    0,   38,    0,    0,    0,    0,    5,
    6,    7,    4,    0,    0,    8,    0,  292,  294,  293,
  295,  296,  211,   46,  268,  277,    0,    0,    0,   43,
   43,   43,   43,   43,   57,   58,    0,    0,    0,    0,
  192,    0,    0,    0,   43,   40,   10,  198,    0,   10,
    0,   25,    0,  133,    0,  134,    0,    0,  131,    0,
    0,    0,    0,    0,  132,    0,  261,    0,    0,    0,
    0,    0,   35,    0,   37,    0,    0,    8,    0,    0,
  195,    0,   43,   12,    0,  120,    0,    0,  130,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  122,    8,
    8,    8,    0,    0,  123,    8,   60,    0,    0,    0,
  291,  279,  280,  281,  282,  283,  284,  285,  286,  287,
  288,  289,  290,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  219,  221,  223,    0,    0,    0,    0,    0,
    0,    0,    0,   44,    0,    0,    0,  141,    0,    0,
    0,   43,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  108,   59,   76,    0,    0,    0,    0,   34,   30,
   32,   31,    0,    0,    0,    0,    0,    0,  271,    0,
    0,  270,   20,   39,   79,   44,   78,    0,   86,   85,
    0,    0,    0,  111,    0,    0,    0,  266,  160,    0,
    0,    0,    0,    0,  178,    0,    0,    0,  124,  133,
  134,  131,  132,    0,    0,  118,    0,    0,    0,  119,
    0,  121,    0,   10,   18,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  275,  276,    0,
    0,    0,  202,    0,    0,    0,   45,   48,    0,  136,
    0,  137,    0,    0,  138,  113,    0,    0,    0,    0,
    0,    0,    0,    0,   14,    0,  200,    0,   27,   33,
    0,    0,    0,    0,    0,    0,    0,   63,   82,   81,
    0,  110,    0,    0,  184,    0,  112,    0,    0,  327,
  329,  328,  115,  109,    0,    8,   84,    0,    0,    0,
    0,    0,  181,  177,  190,    0,  180,  139,   43,  105,
    0,    0,    0,    0,    0,  117,   42,   68,    0,    0,
  273,  272,    0,  205,  274,   47,    0,    0,    0,  100,
    0,    0,    0,    0,    0,    0,    0,    0,   75,   71,
    0,   21,   22,   43,  157,   43,   43,  161,    0,   44,
    0,  188,    0,  187,  330,    0,    0,    0,    0,    0,
  309,    0,  317,    0,  324,  311,  106,    0,    0,  128,
  125,  159,  126,  127,  179,    0,    9,    0,  104,    0,
    9,    0,   19,    0,  204,  163,  165,    0,   65,    0,
    0,    0,    0,    0,    0,   43,    0,    0,    0,    0,
    0,  262,    0,  186,    0,    0,  319,  321,    0,    0,
    0,    0,    0,    0,    0,    0,  183,   44,    0,    0,
   44,   66,  102,    9,   70,    0,    0,    0,   44,    0,
   73,   44,   44,    0,    0,    0,    0,   44,  312,    0,
   44,    0,   44,  297,   83,    0,  326,  306,  307,    0,
  318,    0,  308,    0,    0,  316,  315,    0,  323,  322,
  210,    0,  208,   84,   67,    0,   44,    0,    0,    9,
    9,   44,    9,    9,    0,    0,   44,    0,   84,    0,
  313,  265,  263,  298,  264,    0,    0,  300,    0,    0,
    9,    9,   43,    0,   44,    0,   44,   50,   52,    9,
   54,   55,    0,    0,   88,   89,    0,   44,    0,  304,
    9,   69,   56,    0,    0,   93,   94,    9,   53,    0,
   92,   44,    9,    0,  303,   44,   64,    0,   97,   51,
   91,   77,   72,  302,    0,   98,    0,  301,    0,   44,
   96,
};
short yydgoto[] = {                                       1,
   79,  140,  141,  142,  143,  273,   80,  286,  287,  528,
  244,  245,  373,  449,  390,  543,  587,  588,  577,  605,
  606,   82,   83,  636,  559,  560,  596,  254,  171,  422,
  172,  647,   84,   85,  423,  296,   86,  405,   87,  132,
  179,   89,  301,  276,  303,   90,  277,  249,  251,   91,
   92,   93,   94,   95,  180,  307,  308,  291,  395,  407,
  396,  417,  445,  446,  246,  137,  138,   97,  352,  353,
   98,  242,  115,  113,  166,  290,  553,  400,  470,  471,
  548,  401,  472,  473,  517,  402,  474,  475,  476,    5,
    6,    8,  103,  292,  391,    7,  176,  376,  522,  114,
  356,   99,  357,  492,  288,  167,  530,
};
short yysindex[] = {                                      0,
   40,    0,    0,    0,    0,    0, 1098,   85, -203,  -31,
 -187,    0,  251,    0,  251,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, -151,    0,    0,    0,    0,    0,    0,    0,
    0, -181, -138, 1488, 2658,    0,    0,    0,    0, -131,
 2658,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2658, 2658, 2658, 2658,
 2658, 2658, 2658, 2658, 2658, -114,    0, 1358,    0,    0,
    0,    0,   -9,    0,    0,  -22,    2, -249, -197,    0,
 2832,  -36, 2788,   -1,    0, -183, 5237,  -38, 1228,    0,
    0,    0,    0, 2925, 2925,    0,    3,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   25,  -75,  -66,    0,
    0,    0,    0,    0,    0,    0,   56,   32, 2788,   89,
    0,  -11, 3088, 1228,    0,    0,    0,    0, 4966,    0,
   30,    0,   76,    0,  683,    0,  683,  683,    0,  683,
  683,  683,  683,  683,    0,  683,    0, 2008,   -3, -260,
    8,   94,    0,   15,    0,  157,  100,    0,  108,   48,
    0,  105,    0,    0,  131,    0, 2138,  839,    0, -197,
 -131, 2925, 2925, 2925, 2925, 2925,  -30,   94,    0,    0,
    0,    0,   94, 2832,    0,    0,    0, 1488,  188, 2658,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2658, 2658, 2658, 2658, 2658, 2658, 2658,
 2658, 2658, 2658, 2658, 2658, 2658, 2658, 2658, 2658, 2658,
 2658, 2658,    0,    0,    0, 2658, 2658,  138,  141, 2658,
 2658, 2658, 2658,    0, 1228,  195,    3,    0,  193,    3,
  194,    0,  125,  201,   78, 2658,   80,   81, 1228,   95,
  106,    0,    0,    0,  115,  213, 2658,  213,    0,    0,
    0,    0,   84, 2658,   -4, -297,  122, 5021,    0,  196,
    0,    0,    0,    0,    0,    0,    0,  248,    0,    0,
  124, 2925,  126,    0, 2268,  181,  130,    0,    0,   33,
  136,  185, -195,  139,    0,  211,  137, -224,    0,    0,
    0,    0,    0,  142,   15,    0,  215, 2881, -151,    0,
   94,    0,   86,    0,    0,  683, 5074, 3274, 5290, 5326,
 5362,  281,  564,  564,  761,  761,  761,  761,  831,  831,
  416,  416,  683,  683,  683,  683,  151,    0,    0,  152,
 4913,  150,    0, 5237, 5237,  276,    0,    0,  280,    0,
  283,    0,  154,    3,    0,    0,  158, 1488, 1488,  249,
 1488, 1488,  239, 1488,    0,  292,    0,  294,    0,    0,
  166,  189,  168,  170, 2658,  171,  298,    0,    0,    0,
  240,    0,  264,    0,    0, -146,    0, 1618, 5237,    0,
    0,    0,    0,    0,   32,    0,    0, -197, -197,  315,
 -197, -197,    0,    0,    0, 2925,    0,    0,    0,    0,
    0,  321,  252,   97,  324,    0,    0,    0,  327, 2658,
    0,    0, 2658,    0,    0,    0, 2925, 2925, 1488,    0,
  328,  203,  204,  208,  336,  212,  220, -151,    0,    0,
  343,    0,    0,    0,    0,    0,    0,    0, 2658,    0,
 2658,    0, 2925,    0,    0,  273, 2268, -114, 5129,  275,
    0,  278,    0,  279,    0,    0,    0,  227,    0,    0,
    0,    0,    0,    0,    0,  229,    0,   97,    0,  358,
    0,  357,    0, 3088,    0,    0,    0,  237,    0, 1228,
 1228, 1488, 1488, 1228,  248,    0, 1488,  364,  370,  372,
  683,    0, 4805,    0, 2268, 5237,    0,    0,  340, 2268,
 2268,  390, 2398,  389,  350,  400,    0,    0,  331,  248,
    0,    0,    0,    0,    0, 1228,  248,  373,    0,  286,
    0,    0,    0,  121,  130,  290, 1748,    0,    0, 2528,
    0, 1878,    0,    0,    0, 5237,    0,    0,    0,  345,
    0, 5237,    0,  423, 5129,    0,    0,  273,    0,    0,
    0, 2658,    0,    0,    0,  426,    0,  127, 1228,    0,
    0,    0,    0,    0, 1488,  341,    0,  121,    0, 1228,
    0,    0,    0,    0,    0,  308,  340,    0, 5237,  432,
    0,    0,    0,  346,    0,  127,    0,    0,    0,    0,
    0,    0,  349, 1228,    0,    0,  248,    0, -180,    0,
    0,    0,    0,  384, 1228,    0,    0,    0,    0, 1228,
    0,    0,    0, 2268,    0,    0,    0,  391,    0,    0,
    0,    0,    0,    0,  437,    0,  353,    0, 1228,    0,
    0,
};
short yyrindex[] = {                                      0,
  968,    0,    0,    0,    0,    0,  396,    0,    0,    0,
    0,    0,    0,    0, 2980,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   -5,  420,    0,    0,    0,    0, 3036,
   36,   -2,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  420,  420,  420,  420,
  420,  420,  420,  420,  420,    0,    0,  420,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   23, -159,    0,
  -13,  414,  217,    0,    0,  448, -247,   14,  335,    0,
    0,    0,    0, -133,  -81,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  401,    0,
    0,  442, 1702,  396,    0,    0,    0,    0, -233,    0,
  180,    0,    0,    0,  822,    0, 3352, 3408,    0, 3464,
 3520, 3576, 3632, 3688,    0, 3744,    0,  420,    0,  -46,
    0,  725,    0, 3036,    0, -169, -186,    0,    0,    4,
    0, -176,    0,    0,    0,    0, -267,    0,    0, -159,
    0,    0,    0,    0,    0,    0,    0,  -13,    0,    0,
    0,    0,  -13,  -13,    0,    0,    0,  420,    0,  420,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  420,  420,  420,  420,  420,  420,  420,
  420,  420,  420,  420,  420,  420,  420,  420,  420,  420,
  420,  420,    0,    0,    0,  420,  420,    0,    0,  420,
 -276,  420,  420,    0,  387,    0,    0,    0,    0,    0,
    0,    0,   69,    0,    0, -270,    0,    0,  396,    0,
    0,    0,    0,    0,    0,  452,  420,  452,    0,    0,
    0,    0, -155, -262,  442,    0,    0,  362,    0,    0,
 -245,    0,    0,    0,    0,    0,    0,  409,    0,    0,
    0, -286,    0,    0,  420,   26,    0,    0,    0,  442,
    0,  381,    0,    0,    0, -241,    0,  399,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  -13,    0,    0,    0,    0, 3800,    0, 1959, 1570, 1829,
 4752, 4696, 4584, 4640, 4360, 4416, 4472, 4528, 4248, 4304,
 4136, 4192, 3856, 3912, 3968, 4024,    0,    0,    0,    0,
  399,    0,    0,   87,  -34,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  420,  420,    0,
 -227,  420,  454,   -5,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  420, 5182,    0,    0,    0,    0,
    0,    0,    0, -222,    0,  399,    0,  420,    7,    0,
    0,    0,    0,    0,    0,    0,    0, -159, -159,    0,
 -159, -159,    0,    0,    0,    0,    0,    0,    0,    0,
  -18,    0,  529,    0,    0,    0,    0,    0,    0,  420,
    0,    0,  420,    0,    0,    0, -133,  -81,  420,    0,
    0,    0,    0,    0,  402,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  420,    0,
  420,    0,  482,    0,    0, 4858,  420,  485, -119,   13,
    0,  535,    0,  535,    0,    0,    0,    0,  -87,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2089,    0,    0,    0,    0,    0,  396,
  396,  420, -227,  396,    0,    0, -264, 3166,    0, 3222,
 4080,    0,    0,    0,  420, -221,    0,    0,    0,  420,
   50,  539, -154,    0,  541,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  335,    0,  589,    0,    0,
    0,    0,    0,  542,    0,    0,  420,    0,    0,  420,
    0,  420,    0,    0,    0, -124,    0,    0,    0,  425,
    0, -119,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  420,    0,    0,    0,    0,    0,  545,  396,    0,
    0,    0,    0,    0,  420,    0,    0,  542,    0,  396,
    0,    0,    0,    0,    0,    0,    0,    0,  272,    0,
    0,    0,    0,    0,    0,  545,    0,    0,    0,    0,
    0,    0,    0,  387,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  387,    0,    0,    0,    0,  387,
    0,    0,    0,  420,    0,    0,    0,  460,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  387,    0,
    0,
};
short yygindex[] = {                                      0,
  -29,    0,    0,  548,    0,    0,  -61,    0,    0,    0,
 -242,    1,    0,    0,    0,    0,  230,    0,    0,  214,
    0,    0,  385,    0,  222,    0,    0, -172,  419, -190,
    0,    0,    0,    0,    0,    0,  -43,    0,  -64,   -7,
 -168,    0,    0,  644,    0,    0, -164,  392,  404,  -16,
    0,  -35,    0,  -77,  342,  411,    0,  424,  374,  295,
    0, -337,  -42,  365,   18,  583,  577,  -17,    0,  413,
    0,    0,    0,  837, -400,    0,    0,  -58,    0, -511,
    0,  352,    0,  330,  366,    0,    0,  360, -273,    0,
    0,  -70,    0,  445, -204,    0,  -55,  624,  -62,  -19,
   57, -256, -518,    0,    0,    0,    0,
};
#define YYTABLESIZE 5755
short yytable[] = {                                      88,
  128,  130,  358,  118,  174,  165,  479,   81,  129,  563,
  162,  309,  304,  434,  142,  174,  159,  196,  196,  139,
  142,  403,   43,  142,   96,  199,  142,  133,  142,  182,
  142,  201,  201,  139,  128,  189,   88,  195,  193,  247,
  250,  165,  129,  382,  139,  139,  415,  320,  189,  145,
  147,  148,  150,  151,  152,  153,  154,  156,  464,  144,
  146,  131,  149,  142,  104,  168,  266,  155,  615,  268,
  160,  139,  347,  195,  359,  350,  194,  361,  158,   87,
  107,   87,  197,  188,  125,  193,  626,   14,  116,  116,
  383,   88,  196,  196,  141,  161,  248,  248,   80,  182,
  257,  258,  259,  260,  261,  129,  283,  185,  196,   29,
   29,  129,  194,  460,   16,  265,  416,  203,  189,  320,
  156,  193,  644,  199,  415,  158,   88,  126,  280,  193,
  129,  129,  316,  281,  264,  164,  142,  320,  322,  175,
  278,  139,  139,  139,  325,  410,  196,  139,  139,  310,
  275,  310,  182,  297,  128,   87,  198,  129,  129,  278,
  201,  634,  129,  116,  116,   48,  193,   50,  450,  300,
  306,  189,   80,  600,  310,  311,  312,  313,  314,  116,
  207,  129,  326,  142,  564,   29,  321,  166,  617,  177,
   88,  440,  411,  518,  463,  178,  327,  328,  329,  330,
  331,  332,  333,  334,  335,  336,  337,  338,  339,  340,
  341,  342,  343,  344,  345,  324,  325,  116,  346,  278,
  458,  310,  278,  351,  354,  355,  389,  393,  129,  129,
  197,  197,  363,  490,  129,  142,  105,   88,  139,  480,
  481,  557,  483,  484,  141,  428,  518,  104,  544,  139,
   48,   88,  164,  429,    9,  190,  150,  191,  192,  370,
  193,  387,   26,   26,  136,  426,  149,  169,  108,  108,
   13,  114,  114,  157,  529,  106,   13,  399,  194,  194,
  578,   17,  532,   17,  394,  142,  535,  163,  315,  425,
  107,  107,  173,  576,  170,    2,  196,  533,   36,   36,
  199,  199,  243,   13,    3,  197,  197,   48,  163,  164,
  424,    4,   13,  141,   34,   34,   34,  256,   16,  253,
   16,  197,  104,  105,  128,  128,  142,  128,  128,  575,
  128,  451,  129,  129,  101,  129,  129,  101,  129,   13,
  142,  255,  388,  108,  108,  108,  114,  114,  262,  100,
  101,  209,  209,  194,  263,  209,  102,  209,  174,  197,
   88,   88,  114,   88,   88,  107,   88,  278,   36,  194,
  269,  631,  247,  250,   36,  608,  609,  175,  611,  612,
  469,  107,  639,  381,  175,  442,  443,  641,  131,  447,
  279,  131,  270,  271,  272,  128,  622,  623,  393,  486,
  114,  282,   48,  129,  427,  629,  650,  194,  306,  524,
  284,  526,  494,  420,  489,  351,  637,  119,  506,  107,
  408,  175,  285,  640,  293,  209,  209,  209,  643,  248,
  248,   88,  585,  586,  508,  289,  509,  510,  603,  604,
  294,  511,  209,  513,   28,   28,  295,  298,  519,  516,
  551,  554,  325,    9,  348,  394,  498,  349,  128,  128,
  197,  360,  362,  128,  546,  364,  129,  129,  573,  365,
  366,  129,  368,  369,  209,  187,  117,  187,  375,  380,
  209,  120,  121,  122,  123,  124,  545,  371,  187,  187,
  134,  135,   88,   88,   88,   88,   88,  556,  372,   88,
  538,  539,  516,  562,  542,  565,   13,  374,  171,  384,
  171,  171,  386,  187,  171,   14,  512,  392,  397,  540,
  131,  404,  406,  409,  131,  382,  412,  413,   88,  565,
  414,  419,  562,   13,  562,  418,  206,  206,  431,  432,
  206,  128,  206,  435,  436,  437,  439,  200,  438,  129,
  252,  441,  444,  448,  599,  632,  452,  635,  453,  454,
  455,  456,  187,  457,  459,  398,  187,  187,  187,  187,
  187,   88,  108,  109,  110,  111,  112,   88,  187,  607,
  462,  461,   88,  624,  571,  482,  487,  574,   44,  491,
  618,  493,  488,  499,  580,  581,  500,  501,  583,  584,
  502,  503,  613,   49,  592,  504,   88,  593,  507,  595,
  206,  206,  206,  505,  515,  521,  562,   88,  523,  525,
  527,  178,   88,  534,  536,  142,  172,  206,  172,  172,
  537,  547,  172,  602,  317,  318,  319,  550,  610,  552,
  323,   88,  220,  221,  222,  223,  224,  225,  226,  227,
  228,  229,  230,  231,  232,   49,  558,  567,  233,  206,
  415,  234,  235,  628,  236,  206,  568,  237,  570,  238,
  239,  240,  572,  241,  633,  579,  187,  142,  172,  582,
  172,  172,  200,  590,  172,  597,  142,  172,  642,  172,
  172,  598,  645,  172,  601,  619,  614,  621,   49,   49,
  638,  625,  187,  119,  630,  648,  651,  646,  649,    9,
  142,  175,   10,    9,    9,    9,   15,   13,    9,    9,
    9,    9,    9,    9,   13,    9,    9,    9,    9,    9,
    9,    9,  141,    9,    9,    9,    9,    9,    9,    9,
    9,    9,    9,    9,    9,    9,    9,    9,    9,  162,
    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,
    9,    9,    9,    9,    9,    9,    9,    9,  155,    9,
    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,
    9,    9,    9,    9,    9,    9,    9,  230,  231,  232,
    9,    9,  191,  233,  103,   74,  234,  235,  142,  236,
  187,   43,  237,   17,  238,  239,  240,  191,  241,   16,
   90,    9,  305,   95,    9,   99,  379,  616,  620,  627,
  302,  187,  187,  477,    9,    9,  485,  531,  496,  478,
  200,    9,  119,    9,    9,    9,  514,    9,  367,  589,
    9,  497,    9,  377,   44,  495,    9,  187,   44,   44,
   44,  116,  566,   44,   44,   44,   44,   44,   44,  549,
   44,   44,   44,   44,   44,   44,   44,  541,   44,   44,
   44,   44,   44,   44,   44,   44,   44,   44,   44,   44,
   44,   44,   44,   44,  569,  561,   44,   44,   44,   44,
   44,  378,   44,   44,   44,   44,   44,   44,   44,   44,
   44,   44,   44,   44,   44,   44,   44,   44,   44,   44,
   44,   44,   44,   44,   44,   44,   44,   44,   44,   44,
   44,   44,    0,    0,    0,   44,   44,  222,  223,  224,
  225,  226,  227,  228,  229,  230,  231,  232,    0,    0,
    0,  233,    0,    0,  234,  235,   44,  236,    0,   44,
  237,    0,  238,  239,  240,    0,  241,    0,    0,   44,
   44,    0,    0,    0,    0,    0,   44,    0,   44,   44,
   44,    0,   44,    0,    0,   44,    0,   44,    0,    0,
   13,   44,    0,    0,   13,   13,   13,    0,    0,    0,
   13,   13,   13,    0,   13,    0,   13,   13,   13,   13,
   13,   13,   13,    0,   13,   13,   13,   13,   13,   13,
   13,   13,   13,   13,   13,   13,   13,   13,   13,   13,
    0,    0,   13,   13,   13,   13,   13,  200,   13,   13,
   13,   13,   13,   13,   13,   13,    0,    0,   13,    0,
   13,   13,   13,   13,   13,   13,   13,   13,   13,   13,
   13,   13,   13,   13,   13,   13,   13,   13,    0,    0,
  233,   13,   13,  234,  235,    0,  236,    0,    0,  237,
    0,  238,  239,  240,    0,  241,    0,    0,    0,    0,
    0,    0,   13,    0,    0,   13,  215,  215,  215,    0,
  215,    0,  215,    0,    0,   13,   13,  200,    9,  127,
   11,    0,   13,    0,   13,   13,   13,    0,   13,    0,
    0,   13,    0,   13,    0,   20,   21,   13,   22,   23,
   24,   25,   26,   27,   28,   29,   30,   31,  226,  227,
  228,  229,  230,  231,  232,    0,    0,    0,  233,    0,
    0,  234,  235,    0,  236,    0,    0,  237,    0,  238,
  239,  240,    0,  241,    0,    0,   49,  181,    0,    0,
  215,  215,  215,  215,  215,  215,  215,  215,  215,  215,
  215,  215,  215,  215,  215,  215,  215,  215,  215,  215,
  215,  215,  215,  215,  215,  215,  215,  215,  215,  215,
  215,  215,  215,  215,  215,  215,  182,    0,    0,  183,
  228,  229,  230,  231,  232,    0,    0,    0,  233,  215,
  184,  234,  235,    0,  236,  215,    0,  237,    0,  238,
  239,  240,  185,  241,    0,    0,    0,   11,   11,   11,
    0,  186,  305,   11,   11,   11,    0,   11,    0,    0,
   11,   11,   11,   11,   11,   11,    0,   11,   11,   11,
   11,   11,   11,   11,   11,   11,   11,   11,   11,   11,
   11,   11,   11,    0,    0,   11,   11,   11,   11,   11,
    0,   11,   11,   11,   11,   11,   11,   11,   11,    0,
    0,   11,    0,   11,   11,   11,   11,   11,   11,   11,
   11,   11,   11,   11,   11,   11,   11,   11,   11,   11,
   11,    0,    0,    0,   11,   11,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   11,    0,    0,   11,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   11,   11,
    0,    0,    0,    0,    0,   11,    0,   11,   11,   11,
    0,   11,    0,    0,   11,    0,   11,    9,   10,   11,
   11,    0,    0,   12,   13,   14,    0,   15,    0,    0,
   16,   17,   18,   19,   20,   21,    0,   22,   23,   24,
   25,   26,   27,   28,   29,   30,   31,   32,    0,    0,
   33,    0,    0,    0,    0,    0,   34,   35,   36,   37,
    0,   38,   39,   40,   41,   42,   43,   44,   45,    0,
    0,   46,    0,   47,   48,   49,   50,   51,   52,   53,
   54,   55,   56,   57,   58,   59,   60,   61,   62,   63,
   64,    0,    0,    0,   65,   66,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   67,    0,    0,   68,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   69,   70,
    0,    0,    0,    0,    0,   71,    0,   72,   73,   74,
    0,   75,    0,    0,   76,    0,   77,    9,   10,   11,
   78,    0,    0,   12,   13,   14,    0,   15,    0,    0,
   16,   17,   18,   19,   20,   21,    0,   22,   23,   24,
   25,   26,   27,   28,   29,   30,   31,   32,    0,    0,
   33,    0,    0,    0,    0,    0,   34,   35,   36,   37,
    0,   38,   39,   40,   41,   42,   43,   44,   45,    0,
    0,   46,    0,   47,   48,   49,   50,    0,    0,   53,
   54,   55,   56,   57,   58,   59,   60,   61,   62,   63,
   64,    0,    0,    0,   65,   66,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   67,    0,    0,   68,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   69,   70,
    0,    0,    0,    0,    0,   71,    0,   72,   73,   74,
    0,   75,    0,    0,   76,    0,   77,    9,  127,   11,
   78,    0,    0,    0,   13,   14,    0,   15,    0,    0,
   16,   17,   18,   19,   20,   21,    0,   22,   23,   24,
   25,   26,   27,   28,   29,   30,   31,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   34,   35,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   45,    0,
    0,    0,    0,    0,   48,   49,   50,    0,    0,   53,
   54,   55,   56,   57,   58,   59,   60,   61,   62,   63,
   64,    0,    0,    0,   65,   66,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   67,    0,    0,   68,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   69,   70,
    0,    0,    0,    0,    0,   71,    0,   72,   73,   74,
    0,   75,    0,    0,  158,    0,   77,    9,  127,   11,
   78,    0,    0,    0,   13,    0,    0,   15,    0,    0,
   16,   17,   18,   19,   20,   21,    0,   22,   23,   24,
   25,   26,   27,   28,   29,   30,   31,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   34,   35,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   45,    0,
    0,    0,    0,    0,   48,   49,   50,    0,    0,   53,
   54,   55,   56,   57,   58,   59,   60,   61,   62,   63,
   64,    0,    0,    0,   65,   66,    0,    0,    0,    0,
    0,    0,    0,    0,  238,  238,    0,    0,  238,    0,
  238,    0,    0,    0,    0,   67,    0,    0,   68,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   69,   70,
    0,    0,    0,    0,    0,   71,    0,   72,   73,   74,
    0,   75,    0,    0,   76,    0,   77,    9,  127,   11,
   78,    0,    0,    0,   13,  398,  465,   15,    0,    0,
    0,    0,    0,    0,   20,   21,    0,   22,   23,   24,
   25,   26,   27,   28,   29,   30,   31,    0,  238,  238,
  238,  238,  238,  238,  238,  238,  238,  238,  238,  238,
  238,  238,  238,  238,  238,  238,  238,  238,   45,    0,
    0,    0,    0,    0,  466,   49,   50,    0,    0,   53,
   54,   55,   56,   57,   58,   59,   60,   61,   62,   63,
   64,    0,    0,    0,   65,   66,  467,  238,    0,    0,
    0,    0,    0,  238,    0,    0,  224,  224,    0,    0,
  224,    0,  224,    0,    0,   67,    0,    0,   68,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   69,   70,
    0,    0,    0,    0,    0,   71,    0,   72,   73,   74,
    0,   75,    0,    0,  468,    0,   77,    9,  127,   11,
   78,    0,    0,    0,   13,    0,  591,   15,    0,    0,
    0,    0,    0,    0,   20,   21,    0,   22,   23,   24,
   25,   26,   27,   28,   29,   30,   31,    0,    0,    0,
  224,  224,  224,  224,  224,  224,  224,  224,  224,  224,
  224,  224,  224,  224,  224,  224,    0,  224,   45,    0,
    0,    0,    0,    0,   48,   49,   50,    0,    0,   53,
   54,   55,   56,   57,   58,   59,   60,   61,   62,   63,
   64,    0,    0,    0,   65,   66,  467,    0,    0,  224,
    0,    0,    0,  237,  237,  224,    0,  237,    0,  237,
    0,    0,    0,    0,    0,   67,    0,    0,   68,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   69,   70,
    0,    0,    0,    0,    0,   71,    0,   72,   73,   74,
    0,   75,    0,    0,   76,    0,   77,    9,  127,   11,
   78,    0,    0,    0,   13,  398,  594,   15,    0,    0,
    0,    0,    0,    0,   20,   21,    0,   22,   23,   24,
   25,   26,   27,   28,   29,   30,   31,  237,  237,  237,
  237,  237,  237,  237,  237,  237,  237,  237,  237,  237,
  237,  237,  237,  237,  237,  237,  237,  237,   45,    0,
    0,    0,    0,    0,   48,   49,   50,    0,    0,   53,
   54,   55,   56,   57,   58,   59,   60,   61,   62,   63,
   64,    0,    0,    0,   65,   66,  237,    0,    0,    0,
    0,    0,  237,  239,  239,    0,    0,  239,    0,  239,
    0,    0,    0,    0,    0,   67,    0,    0,   68,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   69,   70,
    0,    0,    0,    0,    0,   71,    0,   72,   73,   74,
    0,   75,    0,    0,  468,    0,   77,    9,  127,   11,
   78,    0,    0,    0,   13,    0,    0,   15,    0,    0,
    0,    0,    0,    0,   20,   21,    0,   22,   23,   24,
   25,   26,   27,   28,   29,   30,   31,  239,  239,  239,
  239,  239,  239,  239,  239,  239,  239,  239,  239,  239,
  239,  239,  239,  239,  239,  239,    0,    0,   45,    0,
    0,    0,    0,    0,   48,   49,   50,    0,    0,   53,
   54,   55,   56,   57,   58,   59,   60,   61,   62,   63,
   64,    0,    0,    0,   65,   66,  239,    0,    0,    0,
    0,    0,  239,  234,  234,    0,    0,  234,    0,  234,
    0,    0,    0,    0,    0,   67,    0,    0,   68,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   69,  274,
    0,    0,    0,    0,    0,   71,    0,   72,   73,   74,
    0,   75,    0,    0,   76,  157,   77,    9,  127,   11,
   78,    0,    0,    0,   13,    0,    0,   15,  299,    0,
    0,    0,    0,    0,   20,   21,    0,   22,   23,   24,
   25,   26,   27,   28,   29,   30,   31,  234,  234,  234,
  234,  234,  234,  234,  234,  234,  234,  234,  234,  234,
  234,  234,  234,    0,  234,    0,    0,    0,   45,    0,
    0,    0,    0,    0,   48,   49,   50,    0,    0,   53,
   54,   55,   56,   57,   58,   59,   60,   61,   62,   63,
   64,    0,    0,    0,   65,   66,  234,    0,    0,    0,
    0,    0,  234,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   67,    0,    0,   68,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   69,  274,
    0,    0,    0,    0,    0,   71,    0,   72,   73,   74,
    0,   75,    0,    0,   76,    0,   77,    9,  127,   11,
   78,    0,    0,    0,   13,  398,    0,   15,    0,    0,
    0,    0,    0,    0,   20,   21,    0,   22,   23,   24,
   25,   26,   27,   28,   29,   30,   31,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   45,    0,
    0,    0,    0,    0,   48,   49,   50,    0,    0,   53,
   54,   55,   56,   57,   58,   59,   60,   61,   62,   63,
   64,    0,    0,    0,   65,   66,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   67,    0,    0,   68,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   69,   70,
    0,    0,    0,    0,    0,   71,    0,   72,   73,   74,
    0,   75,    0,    0,   76,    0,   77,    9,  127,   11,
   78,    0,    0,    0,   13,    0,    0,   15,    0,    0,
    0,    0,    0,    0,   20,   21,    0,   22,   23,   24,
   25,   26,   27,   28,   29,   30,   31,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   45,    0,
    0,    0,    0,    0,   48,   49,   50,    0,    0,   53,
   54,   55,   56,   57,   58,   59,   60,   61,   62,   63,
   64,    0,    0,    0,   65,   66,  467,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   67,    0,    0,   68,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   69,   70,
    0,    0,    0,    0,    0,   71,    0,   72,   73,   74,
    0,   75,    0,    0,   76,    0,   77,    9,  127,   11,
   78,    0,    0,    0,   13,  398,    0,   15,    0,    0,
    0,    0,    0,    0,   20,   21,    0,   22,   23,   24,
   25,   26,   27,   28,   29,   30,   31,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   45,    0,
    0,    0,    0,    0,   48,   49,   50,    0,    0,   53,
   54,   55,   56,   57,   58,   59,   60,   61,   62,   63,
   64,    0,    0,    0,   65,   66,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   67,    0,    0,   68,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   69,   70,
    0,    0,    0,    0,    0,   71,    0,   72,   73,   74,
    0,   75,    0,    0,  468,    0,   77,    9,  127,   11,
   78,    0,    0,    0,   13,    0,    0,   15,    0,    0,
    0,    0,    0,    0,   20,   21,    0,   22,   23,   24,
   25,   26,   27,   28,   29,   30,   31,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   45,    0,
    0,    0,    0,    0,   48,   49,   50,    0,    0,   53,
   54,   55,   56,   57,   58,   59,   60,   61,   62,   63,
   64,    0,    0,    0,   65,   66,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   67,    0,    0,   68,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   69,   70,
    0,    0,    0,    0,    0,   71,    0,   72,   73,   74,
    0,   75,    0,    0,   76,    0,   77,    9,  127,   11,
   78,    0,  174,    0,    0,    0,    0,    0,    0,    0,
   16,   17,   18,   19,   20,   21,    0,   22,   23,   24,
   25,   26,   27,   28,   29,   30,   31,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    9,  127,   11,    0,    0,  174,    0,    0,    0,
    0,    0,    0,    0,    0,   49,  181,    0,   20,   21,
    0,   22,   23,   24,   25,   26,   27,   28,   29,   30,
   31,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    9,  127,   11,    0,    0,  182,    0,    0,  183,   49,
  181,    0,    0,    0,    0,    0,    0,   20,   21,  184,
   22,   23,   24,   25,   26,   27,   28,   29,   30,   31,
    0,  185,    0,    0,    0,    0,    0,    0,    0,    0,
  186,    0,    0,    0,    9,  127,   11,    0,    0,  182,
    0,    0,  183,    0,    0,    0,    0,  420,  421,  181,
    0,   20,   21,  184,   22,   23,   24,   25,   26,   27,
   28,   29,   30,   31,    0,  185,    0,    0,    0,    0,
    0,    0,    0,    0,  186,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  182,    0,
    0,  183,   49,  181,  278,  278,  278,    0,  278,    0,
  278,    0,  184,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  185,    0,    0,    0,    0,    0,
    0,    0,    0,  186,    0,    0,    0,    0,    0,    0,
    0,    0,  182,    0,    0,  183,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  184,    0,    0,    0,
   42,   42,   42,    0,   42,    0,   42,    0,  185,    0,
    0,    0,    0,    0,    0,    0,    0,  186,  278,  278,
  278,  278,  278,  278,  278,  278,  278,  278,  278,  278,
  278,  278,  278,  278,  278,  278,  278,  278,  278,  278,
  278,  278,  278,  278,  278,  278,  278,  278,  278,  278,
  278,  278,  278,  278,  200,    0,    0,  278,    0,    0,
  278,  278,    0,  278,    0,    0,  278,  278,  278,  278,
  278,    0,  278,  278,   42,   42,   42,   42,   42,   42,
   42,   42,   42,   42,   42,   42,   42,   42,   42,   42,
   42,   42,   42,   42,   42,   42,   42,   42,   42,   42,
   42,   42,   42,   42,   42,   42,   42,   42,   42,   42,
    0,    0,    0,   42,    0,    0,   42,   42,    0,   42,
    0,    0,   42,   42,   42,   42,   42,    0,   42,   42,
  314,  314,  314,    0,  314,    0,  314,    0,    0,    0,
    0,    0,  214,    0,  215,  216,  217,  218,  219,  220,
  221,  222,  223,  224,  225,  226,  227,  228,  229,  230,
  231,  232,    0,    0,    0,  233,    0,    0,  234,  235,
    0,  236,    0,    0,  237,    0,  238,  239,  240,    0,
  241,    0,    0,    0,    0,    0,  299,  299,  299,    0,
  299,    0,  299,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  314,  314,  314,  314,  314,  314,
  314,  314,  314,  314,  314,  314,  314,  314,  314,  314,
  314,  314,  314,  314,  314,  314,  314,  314,  314,  314,
  314,  314,  314,  314,  314,  314,  314,  314,  314,  314,
  200,    0,    0,  314,    0,    0,  314,  314,    0,  314,
    0,    0,  314,  314,  314,  314,  314,    0,  314,  314,
  299,  299,  299,  299,  299,  299,  299,  299,  299,  299,
  299,  299,  299,  299,  299,  299,  299,  299,  299,  299,
  299,  299,  299,  299,  299,  299,  299,  299,  299,  299,
  299,  299,  299,  299,  299,  299,    0,    0,    0,  299,
    0,    0,  299,  299,    0,  299,    0,    0,  299,  299,
  299,  299,  299,    0,  299,  299,  214,  214,  214,    0,
  214,    0,  214,    0,    0,    0,    0,    0,    0,    0,
    0,  216,  217,  218,  219,  220,  221,  222,  223,  224,
  225,  226,  227,  228,  229,  230,  231,  232,    0,    0,
    0,  233,    0,    0,  234,  235,    0,  236,    0,    0,
  237,    0,  238,  239,  240,    0,  241,    0,    0,    0,
    0,    0,  216,  216,  216,    0,  216,    0,  216,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  214,  214,  214,  214,  214,  214,  214,  214,  214,  214,
  214,  214,  214,  214,  214,  214,  214,  214,  214,  214,
  214,  214,  214,  214,  214,  214,  214,  214,  214,  214,
  214,  214,  214,  214,  214,  214,    0,    0,  212,  212,
  212,    0,  212,    0,  212,    0,    0,    0,    0,  214,
    0,    0,    0,    0,    0,  214,  216,  216,  216,  216,
  216,  216,  216,  216,  216,  216,  216,  216,  216,  216,
  216,  216,  216,  216,  216,  216,  216,  216,  216,  216,
  216,  216,  216,  216,  216,  216,  216,  216,  216,  216,
  216,  216,    0,    0,  218,  218,  218,    0,  218,    0,
  218,    0,    0,    0,    0,  216,    0,    0,    0,    0,
    0,  216,  212,  212,  212,  212,  212,  212,  212,  212,
  212,  212,  212,  212,  212,  212,  212,  212,  212,  212,
  212,  212,  212,  212,  212,  212,  212,  212,  212,  212,
  212,  212,  212,  212,  212,  212,  212,  212,    0,    0,
  217,  217,  217,    0,  217,    0,  217,    0,    0,    0,
    0,  212,    0,    0,    0,    0,    0,  212,  218,  218,
  218,  218,  218,  218,  218,  218,  218,  218,  218,  218,
  218,  218,  218,  218,  218,  218,  218,  218,  218,  218,
  218,  218,  218,  218,  218,  218,  218,  218,  218,  218,
  218,  218,  218,  218,    0,    0,  220,  220,  220,    0,
  220,    0,  220,    0,    0,    0,    0,  218,    0,    0,
    0,    0,    0,  218,  217,  217,  217,  217,  217,  217,
  217,  217,  217,  217,  217,  217,  217,  217,  217,  217,
  217,  217,  217,  217,  217,  217,  217,  217,  217,  217,
  217,  217,  217,  217,  217,  217,  217,  217,  217,  217,
    0,    0,  222,  222,  222,    0,  222,    0,  222,    0,
    0,    0,    0,  217,    0,    0,    0,    0,    0,  217,
  220,  220,  220,  220,  220,  220,  220,  220,  220,  220,
  220,  220,  220,  220,  220,  220,  220,  220,  220,  220,
  220,  220,  220,  220,  220,  220,  220,  220,  220,  220,
  220,  220,  220,  220,  220,  220,    0,    0,  213,  213,
  213,    0,  213,    0,  213,    0,    0,    0,    0,  220,
    0,    0,    0,    0,    0,  220,  222,  222,  222,  222,
  222,  222,  222,  222,  222,  222,  222,  222,  222,  222,
  222,  222,  222,  222,  222,  222,  222,  222,  222,  222,
  222,  222,  222,  222,  222,  222,  222,  222,  222,  222,
  222,  222,    0,    0,  230,  230,  230,    0,  230,    0,
  230,    0,    0,    0,    0,  222,    0,    0,    0,    0,
    0,  222,  213,  213,  213,  213,  213,  213,  213,  213,
  213,  213,  213,  213,  213,  213,  213,  213,  213,  213,
  213,  213,  213,  213,  213,  213,  213,  213,  213,  213,
  213,  213,  213,  213,  213,  213,  213,  213,    0,    0,
  227,  227,  227,    0,  227,    0,  227,    0,    0,    0,
    0,  213,    0,    0,    0,    0,    0,  213,  230,  230,
  230,  230,  230,  230,  230,  230,  230,  230,  230,  230,
  230,  230,  230,  230,  230,  230,  230,  230,  230,  230,
  230,  230,  230,  230,  230,  230,  230,  230,  230,  230,
  230,  230,  230,  230,    0,    0,  228,  228,  228,    0,
  228,    0,  228,    0,    0,    0,    0,  230,    0,    0,
    0,    0,    0,  230,  227,  227,  227,  227,  227,  227,
  227,  227,  227,  227,  227,  227,  227,  227,  227,  227,
  227,  227,  227,  227,  227,  227,  227,  227,  227,  227,
  227,  227,  227,  227,  227,  227,  227,  227,  227,  227,
    0,    0,  229,  229,  229,    0,  229,    0,  229,    0,
    0,    0,    0,  227,    0,    0,    0,    0,    0,  227,
  228,  228,  228,  228,  228,  228,  228,  228,  228,  228,
  228,  228,  228,  228,  228,  228,  228,  228,  228,  228,
  228,  228,  228,  228,  228,  228,  228,  228,  228,  228,
  228,  228,  228,  228,  228,  228,    0,    0,  231,  231,
  231,    0,  231,    0,  231,    0,    0,    0,    0,  228,
    0,    0,    0,    0,    0,  228,  229,  229,  229,  229,
  229,  229,  229,  229,  229,  229,  229,  229,  229,  229,
  229,  229,  229,  229,  229,  229,  229,  229,  229,  229,
  229,  229,  229,  229,  229,  229,  229,  229,  229,  229,
  229,  229,    0,    0,  267,  267,  267,    0,  267,    0,
  267,    0,    0,    0,    0,  229,    0,    0,    0,    0,
    0,  229,  231,  231,  231,  231,  231,  231,  231,  231,
  231,  231,  231,  231,  231,  231,  231,  231,  231,  231,
  231,  231,  231,  231,  231,  231,  231,  231,  231,  231,
  231,  231,  231,  231,  231,  231,  231,  231,    0,    0,
  225,  225,    0,    0,  225,    0,  225,    0,    0,    0,
    0,  231,    0,    0,    0,    0,    0,  231,  267,  267,
  267,  267,  267,  267,  267,  267,  267,  267,  267,  267,
  267,  267,  267,  267,  267,  267,  267,  267,  267,  267,
  267,  267,  267,  267,  267,  267,  267,  267,  267,  267,
  267,  267,  267,  267,    0,    0,  226,  226,    0,    0,
  226,    0,  226,    0,    0,    0,    0,  267,    0,    0,
    0,    0,    0,  267,  225,  225,  225,  225,  225,  225,
  225,  225,  225,  225,  225,  225,  225,  225,  225,  225,
  225,  225,  225,  225,  225,  225,  225,  225,  225,  225,
  225,  225,  225,  225,  225,  225,  225,    0,    0,    0,
    0,    0,  232,  232,    0,    0,  232,    0,  232,    0,
    0,    0,    0,  225,    0,    0,    0,    0,    0,  225,
  226,  226,  226,  226,  226,  226,  226,  226,  226,  226,
  226,  226,  226,  226,  226,  226,  226,  226,  226,  226,
  226,  226,  226,  226,  226,  226,  226,  226,  226,  226,
  226,  226,  226,    0,    0,    0,    0,    0,  233,  233,
    0,    0,  233,    0,  233,    0,    0,    0,    0,  226,
    0,    0,    0,    0,    0,  226,  232,  232,  232,  232,
  232,  232,  232,  232,  232,  232,  232,  232,  232,  232,
  232,  232,  232,  232,  232,  232,  232,  232,  232,  232,
  232,  232,  232,  232,  232,  232,  232,    0,    0,    0,
    0,    0,    0,    0,  242,  242,    0,    0,  242,    0,
  242,    0,    0,    0,    0,  232,    0,    0,    0,    0,
    0,  232,  233,  233,  233,  233,  233,  233,  233,  233,
  233,  233,  233,  233,  233,  233,  233,  233,  233,  233,
  233,  233,  233,  233,  233,  233,  233,  233,  233,  233,
  233,  233,  233,    0,    0,    0,    0,    0,    0,    0,
  243,  243,    0,    0,  243,    0,  243,    0,    0,    0,
    0,  233,    0,    0,    0,    0,    0,  233,  242,  242,
  242,  242,  242,  242,  242,  242,  242,  242,  242,  242,
  242,  242,  242,  242,  242,  242,  242,  242,  242,  242,
  242,  242,  242,  242,  242,  242,  242,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  244,  244,    0,    0,
  244,    0,  244,    0,    0,    0,    0,  242,    0,    0,
    0,    0,    0,  242,  243,  243,  243,  243,  243,  243,
  243,  243,  243,  243,  243,  243,  243,  243,  243,  243,
  243,  243,  243,  243,  243,  243,  243,  243,  243,  243,
  243,  243,  243,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  245,  245,    0,    0,  245,    0,  245,    0,
    0,    0,    0,  243,    0,    0,    0,    0,    0,  243,
  244,  244,  244,  244,  244,  244,  244,  244,  244,  244,
  244,  244,  244,  244,  244,  244,  244,  244,  244,  244,
  244,  244,  244,  244,  244,  244,  244,  244,  244,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  240,  240,
    0,    0,  240,    0,  240,    0,    0,    0,    0,  244,
    0,    0,    0,    0,    0,  244,  245,  245,  245,  245,
  245,  245,  245,  245,  245,  245,  245,  245,  245,  245,
  245,  245,  245,  245,  245,  245,  245,  245,  245,  245,
  245,  245,  245,  245,  245,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  241,  241,    0,    0,  241,    0,
  241,    0,    0,    0,    0,  245,    0,    0,    0,    0,
    0,  245,  240,  240,  240,  240,  240,  240,  240,  240,
  240,  240,  240,  240,  240,  240,  240,  240,  240,  240,
  240,  240,  240,  240,  240,  240,  240,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  236,  236,    0,    0,  236,    0,  236,    0,    0,    0,
    0,  240,    0,    0,    0,    0,    0,  240,  241,  241,
  241,  241,  241,  241,  241,  241,  241,  241,  241,  241,
  241,  241,  241,  241,  241,  241,  241,  241,  241,  241,
  241,  241,  241,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  235,  235,    0,    0,
  235,    0,  235,    0,    0,    0,    0,  241,    0,    0,
    0,    0,    0,  241,  236,  236,  236,  236,  236,  236,
  236,  236,  236,  236,  236,  236,  236,  236,  236,  236,
  236,  236,  236,  236,  236,  236,  236,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  555,  200,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  236,    0,    0,    0,    0,    0,  236,
  235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
  235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
  235,  235,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   41,    0,   41,    0,   41,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  235,
    0,    0,    0,    0,    0,  235,  201,  202,  203,  204,
  205,  206,  207,  208,  209,  210,  211,  212,  213,  214,
    0,  215,  216,  217,  218,  219,  220,  221,  222,  223,
  224,  225,  226,  227,  228,  229,  230,  231,  232,  200,
    0,    0,  233,  415,    0,  234,  235,    0,  236,    0,
    0,  237,    0,  238,  239,  240,   41,  241,   41,    0,
   41,   41,   41,   41,   41,   41,   41,   41,   41,   41,
   41,   41,   41,    0,   41,   41,   41,   41,   41,   41,
   41,   41,   41,   41,   41,   41,   41,   41,   41,   41,
   41,   41,  200,    0,    0,   41,    0,    0,   41,   41,
    0,   41,    0,    0,   41,    0,   41,   41,   41,    0,
   41,    0,    0,  433,  201,  202,  203,  204,  205,  206,
  207,  208,  209,  210,  211,  212,  213,  214,    0,  215,
  216,  217,  218,  219,  220,  221,  222,  223,  224,  225,
  226,  227,  228,  229,  230,  231,  232,  200,    0,    0,
  233,    0,    0,  234,  235,    0,  236,    0,    0,  237,
    0,  238,  239,  240,    0,  241,  267,  201,  202,  203,
  204,  205,  206,  207,  208,  209,  210,  211,  212,  213,
  214,    0,  215,  216,  217,  218,  219,  220,  221,  222,
  223,  224,  225,  226,  227,  228,  229,  230,  231,  232,
  200,    0,    0,  233,    0,    0,  234,  235,    0,  236,
    0,    0,  237,    0,  238,  239,  240,    0,  241,    0,
    0,  385,  201,  202,  203,  204,  205,  206,  207,  208,
  209,  210,  211,  212,  213,  214,    0,  215,  216,  217,
  218,  219,  220,  221,  222,  223,  224,  225,  226,  227,
  228,  229,  230,  231,  232,  200,    0,    0,  233,    0,
    0,  234,  235,    0,  236,    0,    0,  237,    0,  238,
  239,  240,    0,  241,    0,  201,  202,  203,  204,  205,
  206,  207,  208,  209,  210,  211,  212,  213,  214,  430,
  215,  216,  217,  218,  219,  220,  221,  222,  223,  224,
  225,  226,  227,  228,  229,  230,  231,  232,  266,    0,
    0,  233,    0,    0,  234,  235,    0,  236,    0,    0,
  237,    0,  238,  239,  240,    0,  241,  520,    0,    0,
  201,  202,  203,  204,  205,  206,  207,  208,  209,  210,
  211,  212,  213,  214,    0,  215,  216,  217,  218,  219,
  220,  221,  222,  223,  224,  225,  226,  227,  228,  229,
  230,  231,  232,  200,    0,    0,  233,    0,    0,  234,
  235,    0,  236,    0,    0,  237,    0,  238,  239,  240,
    0,  241,  266,  266,  266,  266,  266,  266,  266,  266,
  266,  266,  266,  266,  266,  266,  266,    0,  266,  266,
  266,  266,  266,  266,  266,  266,  266,  266,  266,  266,
  266,  266,  266,  266,  266,  266,  200,    0,    0,  266,
    0,    0,  266,  266,    0,  266,    0,    0,  266,    0,
  266,  266,  266,    0,  266,    0,    0,    0,  201,  202,
  203,  204,  205,  206,  207,  208,  209,  210,  211,  212,
  213,  214,  200,  215,  216,  217,  218,  219,  220,  221,
  222,  223,  224,  225,  226,  227,  228,  229,  230,  231,
  232,    0,    0,    0,  233,    0,    0,  234,  235,    0,
  236,    0,    0,  237,    0,  238,  239,  240,  200,  241,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  217,  218,
  219,  220,  221,  222,  223,  224,  225,  226,  227,  228,
  229,  230,  231,  232,    0,    0,    0,  233,    0,    0,
  234,  235,    0,  236,    0,    0,  237,    0,  238,  239,
  240,    0,  241,    0,    0,  218,  219,  220,  221,  222,
  223,  224,  225,  226,  227,  228,  229,  230,  231,  232,
    0,    0,    0,  233,    0,    0,  234,  235,    0,  236,
    0,    0,  237,    0,  238,  239,  240,    0,  241,    0,
    0,    0,  219,  220,  221,  222,  223,  224,  225,  226,
  227,  228,  229,  230,  231,  232,    0,    0,    0,  233,
    0,    0,  234,  235,    0,  236,    0,    0,  237,    0,
  238,  239,  240,    0,  241,
};
short yycheck[] = {                                       7,
   44,   44,  245,   33,  265,   83,  407,    7,   44,  521,
   81,  180,  177,  351,  291,  265,   78,  265,  266,  265,
  291,  295,  268,  291,    7,   96,  291,   45,  291,  271,
  317,  265,  266,   51,   78,   91,   44,   93,  266,  104,
  105,  119,   78,  341,  290,  291,  271,  269,  271,   67,
   68,   69,   70,   71,   72,   73,   74,   75,  396,   67,
   68,   44,   70,  291,  268,   85,  137,   75,  587,  140,
   78,  317,  237,  129,  247,  240,   93,  250,  341,  266,
  268,  268,  266,   91,  266,   93,  605,  268,  265,  266,
  388,   99,  340,  341,  317,   78,  104,  105,  268,  341,
  120,  121,  122,  123,  124,  265,  162,  394,  356,  265,
  266,  271,  129,  387,  269,  135,  341,  394,  341,  341,
  388,  129,  634,  394,  271,  388,  134,  266,  389,  394,
  290,  291,  188,  394,  134,  269,  291,  193,  194,  389,
  158,  387,  388,  389,  269,  341,  394,  393,  394,  269,
  158,  271,  394,  173,  198,  342,  340,  317,  318,  177,
  394,  342,  198,  340,  341,  317,  394,  319,  373,  177,
  178,  394,  342,  574,  182,  183,  184,  185,  186,  356,
  268,  341,  200,  317,  522,  341,  194,  269,  589,  387,
  198,  364,  388,  467,  341,  393,  214,  215,  216,  217,
  218,  219,  220,  221,  222,  223,  224,  225,  226,  227,
  228,  229,  230,  231,  232,  198,  341,  394,  236,  237,
  385,  341,  240,  241,  242,  243,  288,  292,  388,  389,
  265,  266,  252,  424,  394,  317,  268,  245,  256,  408,
  409,  515,  411,  412,  291,  323,  520,  266,  505,  267,
  317,  259,  319,  324,  342,  292,  274,  294,  295,  259,
  266,  281,  265,  266,  396,  321,  274,  290,  265,  266,
  317,  265,  266,  388,  479,  307,  290,  295,  265,  266,
  537,  269,  487,  271,  292,  291,  491,  318,  319,  319,
  265,  266,  291,  536,  317,  256,  298,  488,  317,  318,
  265,  266,  341,  317,  265,  340,  341,  317,  318,  319,
  318,  272,  290,  291,  317,  318,  319,  393,  269,  317,
  271,  356,  341,  268,  368,  369,  291,  371,  372,  534,
  374,  374,  368,  369,  266,  371,  372,  269,  374,  317,
  291,  317,  286,  340,  341,  342,  340,  341,  317,  265,
  266,  265,  266,  340,  266,  269,  272,  271,  265,  394,
  368,  369,  356,  371,  372,  340,  374,  385,  387,  356,
  341,  614,  437,  438,  393,  580,  581,  389,  583,  584,
  398,  356,  625,  388,  389,  368,  369,  630,  371,  372,
  394,  374,  317,  318,  319,  439,  601,  602,  463,  419,
  394,  394,  317,  439,  319,  610,  649,  394,  416,  472,
  396,  474,  430,  317,  318,  433,  621,   33,  448,  394,
  388,  389,  266,  628,  317,  339,  340,  341,  633,  437,
  438,  439,  312,  313,  454,  336,  456,  457,  312,  313,
  393,  459,  356,  461,  265,  266,  342,  317,  468,  467,
  509,  510,  265,    0,  317,  463,  439,  317,  502,  503,
  266,  269,  269,  507,  507,  341,  502,  503,  530,  269,
  393,  507,  393,  393,  388,   91,   32,   93,  266,  396,
  394,   37,   38,   39,   40,   41,  506,  393,  104,  105,
   46,   47,  500,  501,  502,  503,  504,  515,  393,  507,
  500,  501,  520,  521,  504,  523,  290,  393,  292,  388,
  294,  295,  317,  129,  298,  268,  460,  394,  393,  502,
  503,  341,  393,  388,  507,  341,  388,  317,  536,  547,
  394,  317,  550,  317,  552,  394,  265,  266,  388,  388,
  269,  585,  271,  394,  269,  266,  393,  267,  266,  585,
  106,  394,  304,  315,  572,  617,  265,  619,  265,  394,
  372,  394,  178,  394,  394,  268,  182,  183,  184,  185,
  186,  579,  322,  323,  324,  325,  326,  585,  194,  579,
  317,  342,  590,  603,  528,  271,  266,  531,    0,  266,
  590,  265,  341,  266,  538,  539,  394,  394,  542,  543,
  393,  266,  585,  269,  548,  394,  614,  551,  266,  553,
  339,  340,  341,  394,  342,  341,  634,  625,  341,  341,
  394,  393,  630,  266,  268,  291,  292,  356,  294,  295,
  394,  268,  298,  577,  190,  191,  192,  268,  582,  268,
  196,  649,  362,  363,  364,  365,  366,  367,  368,  369,
  370,  371,  372,  373,  374,  269,  317,  269,  378,  388,
  271,  381,  382,  607,  384,  394,  317,  387,  269,  389,
  390,  391,  342,  393,  618,  303,  292,  291,  292,  394,
  294,  295,  267,  394,  298,  341,  291,  292,  632,  294,
  295,  269,  636,  298,  269,  388,  356,  266,  312,  313,
  317,  356,  318,  319,  356,  269,  650,  317,  356,  256,
  291,  298,  265,  260,  261,  262,  265,  317,  265,  266,
  267,  268,  269,  270,    0,  272,  273,  274,  275,  276,
  277,  278,  291,  280,  281,  282,  283,  284,  285,  286,
  287,  288,  289,  290,  291,  292,  293,  294,  295,  388,
  342,  298,  299,  300,  301,  302,  303,  304,  305,  306,
  307,  308,  309,  310,  311,  312,  313,  314,  388,  316,
  317,  318,  319,  320,  321,  322,  323,  324,  325,  326,
  327,  328,  329,  330,  331,  332,  333,  372,  373,  374,
  337,  338,  394,  378,  266,  394,  381,  382,  317,  384,
  416,  317,  387,  269,  389,  390,  391,  269,  393,  269,
  269,  358,  388,  269,  361,  356,  269,  588,  597,  606,
  177,  437,  438,  405,  371,  372,  416,  486,  437,  406,
  267,  378,  448,  380,  381,  382,  463,  384,  256,  545,
  387,  438,  389,  267,  256,  433,  393,  463,  260,  261,
  262,   15,  523,  265,  266,  267,  268,  269,  270,  508,
  272,  273,  274,  275,  276,  277,  278,  503,  280,  281,
  282,  283,  284,  285,  286,  287,  288,  289,  290,  291,
  292,  293,  294,  295,  525,  520,  298,  299,  300,  301,
  302,  268,  304,  305,  306,  307,  308,  309,  310,  311,
  312,  313,  314,  315,  316,  317,  318,  319,  320,  321,
  322,  323,  324,  325,  326,  327,  328,  329,  330,  331,
  332,  333,   -1,   -1,   -1,  337,  338,  364,  365,  366,
  367,  368,  369,  370,  371,  372,  373,  374,   -1,   -1,
   -1,  378,   -1,   -1,  381,  382,  358,  384,   -1,  361,
  387,   -1,  389,  390,  391,   -1,  393,   -1,   -1,  371,
  372,   -1,   -1,   -1,   -1,   -1,  378,   -1,  380,  381,
  382,   -1,  384,   -1,   -1,  387,   -1,  389,   -1,   -1,
  256,  393,   -1,   -1,  260,  261,  262,   -1,   -1,   -1,
  266,  267,  268,   -1,  270,   -1,  272,  273,  274,  275,
  276,  277,  278,   -1,  280,  281,  282,  283,  284,  285,
  286,  287,  288,  289,  290,  291,  292,  293,  294,  295,
   -1,   -1,  298,  299,  300,  301,  302,  267,  304,  305,
  306,  307,  308,  309,  310,  311,   -1,   -1,  314,   -1,
  316,  317,  318,  319,  320,  321,  322,  323,  324,  325,
  326,  327,  328,  329,  330,  331,  332,  333,   -1,   -1,
  378,  337,  338,  381,  382,   -1,  384,   -1,   -1,  387,
   -1,  389,  390,  391,   -1,  393,   -1,   -1,   -1,   -1,
   -1,   -1,  358,   -1,   -1,  361,  265,  266,  267,   -1,
  269,   -1,  271,   -1,   -1,  371,  372,  267,  260,  261,
  262,   -1,  378,   -1,  380,  381,  382,   -1,  384,   -1,
   -1,  387,   -1,  389,   -1,  277,  278,  393,  280,  281,
  282,  283,  284,  285,  286,  287,  288,  289,  368,  369,
  370,  371,  372,  373,  374,   -1,   -1,   -1,  378,   -1,
   -1,  381,  382,   -1,  384,   -1,   -1,  387,   -1,  389,
  390,  391,   -1,  393,   -1,   -1,  318,  319,   -1,   -1,
  339,  340,  341,  342,  343,  344,  345,  346,  347,  348,
  349,  350,  351,  352,  353,  354,  355,  356,  357,  358,
  359,  360,  361,  362,  363,  364,  365,  366,  367,  368,
  369,  370,  371,  372,  373,  374,  358,   -1,   -1,  361,
  370,  371,  372,  373,  374,   -1,   -1,   -1,  378,  388,
  372,  381,  382,   -1,  384,  394,   -1,  387,   -1,  389,
  390,  391,  384,  393,   -1,   -1,   -1,  260,  261,  262,
   -1,  393,  394,  266,  267,  268,   -1,  270,   -1,   -1,
  273,  274,  275,  276,  277,  278,   -1,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,  290,  291,  292,
  293,  294,  295,   -1,   -1,  298,  299,  300,  301,  302,
   -1,  304,  305,  306,  307,  308,  309,  310,  311,   -1,
   -1,  314,   -1,  316,  317,  318,  319,  320,  321,  322,
  323,  324,  325,  326,  327,  328,  329,  330,  331,  332,
  333,   -1,   -1,   -1,  337,  338,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  358,   -1,   -1,  361,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  371,  372,
   -1,   -1,   -1,   -1,   -1,  378,   -1,  380,  381,  382,
   -1,  384,   -1,   -1,  387,   -1,  389,  260,  261,  262,
  393,   -1,   -1,  266,  267,  268,   -1,  270,   -1,   -1,
  273,  274,  275,  276,  277,  278,   -1,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,  290,   -1,   -1,
  293,   -1,   -1,   -1,   -1,   -1,  299,  300,  301,  302,
   -1,  304,  305,  306,  307,  308,  309,  310,  311,   -1,
   -1,  314,   -1,  316,  317,  318,  319,  320,  321,  322,
  323,  324,  325,  326,  327,  328,  329,  330,  331,  332,
  333,   -1,   -1,   -1,  337,  338,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  358,   -1,   -1,  361,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  371,  372,
   -1,   -1,   -1,   -1,   -1,  378,   -1,  380,  381,  382,
   -1,  384,   -1,   -1,  387,   -1,  389,  260,  261,  262,
  393,   -1,   -1,  266,  267,  268,   -1,  270,   -1,   -1,
  273,  274,  275,  276,  277,  278,   -1,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,  290,   -1,   -1,
  293,   -1,   -1,   -1,   -1,   -1,  299,  300,  301,  302,
   -1,  304,  305,  306,  307,  308,  309,  310,  311,   -1,
   -1,  314,   -1,  316,  317,  318,  319,   -1,   -1,  322,
  323,  324,  325,  326,  327,  328,  329,  330,  331,  332,
  333,   -1,   -1,   -1,  337,  338,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  358,   -1,   -1,  361,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  371,  372,
   -1,   -1,   -1,   -1,   -1,  378,   -1,  380,  381,  382,
   -1,  384,   -1,   -1,  387,   -1,  389,  260,  261,  262,
  393,   -1,   -1,   -1,  267,  268,   -1,  270,   -1,   -1,
  273,  274,  275,  276,  277,  278,   -1,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  299,  300,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  311,   -1,
   -1,   -1,   -1,   -1,  317,  318,  319,   -1,   -1,  322,
  323,  324,  325,  326,  327,  328,  329,  330,  331,  332,
  333,   -1,   -1,   -1,  337,  338,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  358,   -1,   -1,  361,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  371,  372,
   -1,   -1,   -1,   -1,   -1,  378,   -1,  380,  381,  382,
   -1,  384,   -1,   -1,  387,   -1,  389,  260,  261,  262,
  393,   -1,   -1,   -1,  267,   -1,   -1,  270,   -1,   -1,
  273,  274,  275,  276,  277,  278,   -1,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  299,  300,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  311,   -1,
   -1,   -1,   -1,   -1,  317,  318,  319,   -1,   -1,  322,
  323,  324,  325,  326,  327,  328,  329,  330,  331,  332,
  333,   -1,   -1,   -1,  337,  338,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  265,  266,   -1,   -1,  269,   -1,
  271,   -1,   -1,   -1,   -1,  358,   -1,   -1,  361,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  371,  372,
   -1,   -1,   -1,   -1,   -1,  378,   -1,  380,  381,  382,
   -1,  384,   -1,   -1,  387,   -1,  389,  260,  261,  262,
  393,   -1,   -1,   -1,  267,  268,  269,  270,   -1,   -1,
   -1,   -1,   -1,   -1,  277,  278,   -1,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,   -1,  339,  340,
  341,  342,  343,  344,  345,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,  356,  357,  358,  311,   -1,
   -1,   -1,   -1,   -1,  317,  318,  319,   -1,   -1,  322,
  323,  324,  325,  326,  327,  328,  329,  330,  331,  332,
  333,   -1,   -1,   -1,  337,  338,  339,  388,   -1,   -1,
   -1,   -1,   -1,  394,   -1,   -1,  265,  266,   -1,   -1,
  269,   -1,  271,   -1,   -1,  358,   -1,   -1,  361,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  371,  372,
   -1,   -1,   -1,   -1,   -1,  378,   -1,  380,  381,  382,
   -1,  384,   -1,   -1,  387,   -1,  389,  260,  261,  262,
  393,   -1,   -1,   -1,  267,   -1,  269,  270,   -1,   -1,
   -1,   -1,   -1,   -1,  277,  278,   -1,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,   -1,   -1,   -1,
  339,  340,  341,  342,  343,  344,  345,  346,  347,  348,
  349,  350,  351,  352,  353,  354,   -1,  356,  311,   -1,
   -1,   -1,   -1,   -1,  317,  318,  319,   -1,   -1,  322,
  323,  324,  325,  326,  327,  328,  329,  330,  331,  332,
  333,   -1,   -1,   -1,  337,  338,  339,   -1,   -1,  388,
   -1,   -1,   -1,  265,  266,  394,   -1,  269,   -1,  271,
   -1,   -1,   -1,   -1,   -1,  358,   -1,   -1,  361,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  371,  372,
   -1,   -1,   -1,   -1,   -1,  378,   -1,  380,  381,  382,
   -1,  384,   -1,   -1,  387,   -1,  389,  260,  261,  262,
  393,   -1,   -1,   -1,  267,  268,  269,  270,   -1,   -1,
   -1,   -1,   -1,   -1,  277,  278,   -1,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,  339,  340,  341,
  342,  343,  344,  345,  346,  347,  348,  349,  350,  351,
  352,  353,  354,  355,  356,  357,  358,  359,  311,   -1,
   -1,   -1,   -1,   -1,  317,  318,  319,   -1,   -1,  322,
  323,  324,  325,  326,  327,  328,  329,  330,  331,  332,
  333,   -1,   -1,   -1,  337,  338,  388,   -1,   -1,   -1,
   -1,   -1,  394,  265,  266,   -1,   -1,  269,   -1,  271,
   -1,   -1,   -1,   -1,   -1,  358,   -1,   -1,  361,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  371,  372,
   -1,   -1,   -1,   -1,   -1,  378,   -1,  380,  381,  382,
   -1,  384,   -1,   -1,  387,   -1,  389,  260,  261,  262,
  393,   -1,   -1,   -1,  267,   -1,   -1,  270,   -1,   -1,
   -1,   -1,   -1,   -1,  277,  278,   -1,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,  339,  340,  341,
  342,  343,  344,  345,  346,  347,  348,  349,  350,  351,
  352,  353,  354,  355,  356,  357,   -1,   -1,  311,   -1,
   -1,   -1,   -1,   -1,  317,  318,  319,   -1,   -1,  322,
  323,  324,  325,  326,  327,  328,  329,  330,  331,  332,
  333,   -1,   -1,   -1,  337,  338,  388,   -1,   -1,   -1,
   -1,   -1,  394,  265,  266,   -1,   -1,  269,   -1,  271,
   -1,   -1,   -1,   -1,   -1,  358,   -1,   -1,  361,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  371,  372,
   -1,   -1,   -1,   -1,   -1,  378,   -1,  380,  381,  382,
   -1,  384,   -1,   -1,  387,  388,  389,  260,  261,  262,
  393,   -1,   -1,   -1,  267,   -1,   -1,  270,  271,   -1,
   -1,   -1,   -1,   -1,  277,  278,   -1,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,  339,  340,  341,
  342,  343,  344,  345,  346,  347,  348,  349,  350,  351,
  352,  353,  354,   -1,  356,   -1,   -1,   -1,  311,   -1,
   -1,   -1,   -1,   -1,  317,  318,  319,   -1,   -1,  322,
  323,  324,  325,  326,  327,  328,  329,  330,  331,  332,
  333,   -1,   -1,   -1,  337,  338,  388,   -1,   -1,   -1,
   -1,   -1,  394,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  358,   -1,   -1,  361,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  371,  372,
   -1,   -1,   -1,   -1,   -1,  378,   -1,  380,  381,  382,
   -1,  384,   -1,   -1,  387,   -1,  389,  260,  261,  262,
  393,   -1,   -1,   -1,  267,  268,   -1,  270,   -1,   -1,
   -1,   -1,   -1,   -1,  277,  278,   -1,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  311,   -1,
   -1,   -1,   -1,   -1,  317,  318,  319,   -1,   -1,  322,
  323,  324,  325,  326,  327,  328,  329,  330,  331,  332,
  333,   -1,   -1,   -1,  337,  338,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  358,   -1,   -1,  361,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  371,  372,
   -1,   -1,   -1,   -1,   -1,  378,   -1,  380,  381,  382,
   -1,  384,   -1,   -1,  387,   -1,  389,  260,  261,  262,
  393,   -1,   -1,   -1,  267,   -1,   -1,  270,   -1,   -1,
   -1,   -1,   -1,   -1,  277,  278,   -1,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  311,   -1,
   -1,   -1,   -1,   -1,  317,  318,  319,   -1,   -1,  322,
  323,  324,  325,  326,  327,  328,  329,  330,  331,  332,
  333,   -1,   -1,   -1,  337,  338,  339,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  358,   -1,   -1,  361,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  371,  372,
   -1,   -1,   -1,   -1,   -1,  378,   -1,  380,  381,  382,
   -1,  384,   -1,   -1,  387,   -1,  389,  260,  261,  262,
  393,   -1,   -1,   -1,  267,  268,   -1,  270,   -1,   -1,
   -1,   -1,   -1,   -1,  277,  278,   -1,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  311,   -1,
   -1,   -1,   -1,   -1,  317,  318,  319,   -1,   -1,  322,
  323,  324,  325,  326,  327,  328,  329,  330,  331,  332,
  333,   -1,   -1,   -1,  337,  338,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  358,   -1,   -1,  361,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  371,  372,
   -1,   -1,   -1,   -1,   -1,  378,   -1,  380,  381,  382,
   -1,  384,   -1,   -1,  387,   -1,  389,  260,  261,  262,
  393,   -1,   -1,   -1,  267,   -1,   -1,  270,   -1,   -1,
   -1,   -1,   -1,   -1,  277,  278,   -1,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  311,   -1,
   -1,   -1,   -1,   -1,  317,  318,  319,   -1,   -1,  322,
  323,  324,  325,  326,  327,  328,  329,  330,  331,  332,
  333,   -1,   -1,   -1,  337,  338,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  358,   -1,   -1,  361,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  371,  372,
   -1,   -1,   -1,   -1,   -1,  378,   -1,  380,  381,  382,
   -1,  384,   -1,   -1,  387,   -1,  389,  260,  261,  262,
  393,   -1,  265,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  273,  274,  275,  276,  277,  278,   -1,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  260,  261,  262,   -1,   -1,  265,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  318,  319,   -1,  277,  278,
   -1,  280,  281,  282,  283,  284,  285,  286,  287,  288,
  289,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  260,  261,  262,   -1,   -1,  358,   -1,   -1,  361,  318,
  319,   -1,   -1,   -1,   -1,   -1,   -1,  277,  278,  372,
  280,  281,  282,  283,  284,  285,  286,  287,  288,  289,
   -1,  384,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  393,   -1,   -1,   -1,  260,  261,  262,   -1,   -1,  358,
   -1,   -1,  361,   -1,   -1,   -1,   -1,  317,  318,  319,
   -1,  277,  278,  372,  280,  281,  282,  283,  284,  285,
  286,  287,  288,  289,   -1,  384,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  393,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  358,   -1,
   -1,  361,  318,  319,  265,  266,  267,   -1,  269,   -1,
  271,   -1,  372,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  384,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  393,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  358,   -1,   -1,  361,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  372,   -1,   -1,   -1,
  265,  266,  267,   -1,  269,   -1,  271,   -1,  384,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  393,  339,  340,
  341,  342,  343,  344,  345,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,  356,  357,  358,  359,  360,
  361,  362,  363,  364,  365,  366,  367,  368,  369,  370,
  371,  372,  373,  374,  267,   -1,   -1,  378,   -1,   -1,
  381,  382,   -1,  384,   -1,   -1,  387,  388,  389,  390,
  391,   -1,  393,  394,  339,  340,  341,  342,  343,  344,
  345,  346,  347,  348,  349,  350,  351,  352,  353,  354,
  355,  356,  357,  358,  359,  360,  361,  362,  363,  364,
  365,  366,  367,  368,  369,  370,  371,  372,  373,  374,
   -1,   -1,   -1,  378,   -1,   -1,  381,  382,   -1,  384,
   -1,   -1,  387,  388,  389,  390,  391,   -1,  393,  394,
  265,  266,  267,   -1,  269,   -1,  271,   -1,   -1,   -1,
   -1,   -1,  355,   -1,  357,  358,  359,  360,  361,  362,
  363,  364,  365,  366,  367,  368,  369,  370,  371,  372,
  373,  374,   -1,   -1,   -1,  378,   -1,   -1,  381,  382,
   -1,  384,   -1,   -1,  387,   -1,  389,  390,  391,   -1,
  393,   -1,   -1,   -1,   -1,   -1,  265,  266,  267,   -1,
  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  339,  340,  341,  342,  343,  344,
  345,  346,  347,  348,  349,  350,  351,  352,  353,  354,
  355,  356,  357,  358,  359,  360,  361,  362,  363,  364,
  365,  366,  367,  368,  369,  370,  371,  372,  373,  374,
  267,   -1,   -1,  378,   -1,   -1,  381,  382,   -1,  384,
   -1,   -1,  387,  388,  389,  390,  391,   -1,  393,  394,
  339,  340,  341,  342,  343,  344,  345,  346,  347,  348,
  349,  350,  351,  352,  353,  354,  355,  356,  357,  358,
  359,  360,  361,  362,  363,  364,  365,  366,  367,  368,
  369,  370,  371,  372,  373,  374,   -1,   -1,   -1,  378,
   -1,   -1,  381,  382,   -1,  384,   -1,   -1,  387,  388,
  389,  390,  391,   -1,  393,  394,  265,  266,  267,   -1,
  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  358,  359,  360,  361,  362,  363,  364,  365,  366,
  367,  368,  369,  370,  371,  372,  373,  374,   -1,   -1,
   -1,  378,   -1,   -1,  381,  382,   -1,  384,   -1,   -1,
  387,   -1,  389,  390,  391,   -1,  393,   -1,   -1,   -1,
   -1,   -1,  265,  266,  267,   -1,  269,   -1,  271,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  339,  340,  341,  342,  343,  344,  345,  346,  347,  348,
  349,  350,  351,  352,  353,  354,  355,  356,  357,  358,
  359,  360,  361,  362,  363,  364,  365,  366,  367,  368,
  369,  370,  371,  372,  373,  374,   -1,   -1,  265,  266,
  267,   -1,  269,   -1,  271,   -1,   -1,   -1,   -1,  388,
   -1,   -1,   -1,   -1,   -1,  394,  339,  340,  341,  342,
  343,  344,  345,  346,  347,  348,  349,  350,  351,  352,
  353,  354,  355,  356,  357,  358,  359,  360,  361,  362,
  363,  364,  365,  366,  367,  368,  369,  370,  371,  372,
  373,  374,   -1,   -1,  265,  266,  267,   -1,  269,   -1,
  271,   -1,   -1,   -1,   -1,  388,   -1,   -1,   -1,   -1,
   -1,  394,  339,  340,  341,  342,  343,  344,  345,  346,
  347,  348,  349,  350,  351,  352,  353,  354,  355,  356,
  357,  358,  359,  360,  361,  362,  363,  364,  365,  366,
  367,  368,  369,  370,  371,  372,  373,  374,   -1,   -1,
  265,  266,  267,   -1,  269,   -1,  271,   -1,   -1,   -1,
   -1,  388,   -1,   -1,   -1,   -1,   -1,  394,  339,  340,
  341,  342,  343,  344,  345,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,  356,  357,  358,  359,  360,
  361,  362,  363,  364,  365,  366,  367,  368,  369,  370,
  371,  372,  373,  374,   -1,   -1,  265,  266,  267,   -1,
  269,   -1,  271,   -1,   -1,   -1,   -1,  388,   -1,   -1,
   -1,   -1,   -1,  394,  339,  340,  341,  342,  343,  344,
  345,  346,  347,  348,  349,  350,  351,  352,  353,  354,
  355,  356,  357,  358,  359,  360,  361,  362,  363,  364,
  365,  366,  367,  368,  369,  370,  371,  372,  373,  374,
   -1,   -1,  265,  266,  267,   -1,  269,   -1,  271,   -1,
   -1,   -1,   -1,  388,   -1,   -1,   -1,   -1,   -1,  394,
  339,  340,  341,  342,  343,  344,  345,  346,  347,  348,
  349,  350,  351,  352,  353,  354,  355,  356,  357,  358,
  359,  360,  361,  362,  363,  364,  365,  366,  367,  368,
  369,  370,  371,  372,  373,  374,   -1,   -1,  265,  266,
  267,   -1,  269,   -1,  271,   -1,   -1,   -1,   -1,  388,
   -1,   -1,   -1,   -1,   -1,  394,  339,  340,  341,  342,
  343,  344,  345,  346,  347,  348,  349,  350,  351,  352,
  353,  354,  355,  356,  357,  358,  359,  360,  361,  362,
  363,  364,  365,  366,  367,  368,  369,  370,  371,  372,
  373,  374,   -1,   -1,  265,  266,  267,   -1,  269,   -1,
  271,   -1,   -1,   -1,   -1,  388,   -1,   -1,   -1,   -1,
   -1,  394,  339,  340,  341,  342,  343,  344,  345,  346,
  347,  348,  349,  350,  351,  352,  353,  354,  355,  356,
  357,  358,  359,  360,  361,  362,  363,  364,  365,  366,
  367,  368,  369,  370,  371,  372,  373,  374,   -1,   -1,
  265,  266,  267,   -1,  269,   -1,  271,   -1,   -1,   -1,
   -1,  388,   -1,   -1,   -1,   -1,   -1,  394,  339,  340,
  341,  342,  343,  344,  345,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,  356,  357,  358,  359,  360,
  361,  362,  363,  364,  365,  366,  367,  368,  369,  370,
  371,  372,  373,  374,   -1,   -1,  265,  266,  267,   -1,
  269,   -1,  271,   -1,   -1,   -1,   -1,  388,   -1,   -1,
   -1,   -1,   -1,  394,  339,  340,  341,  342,  343,  344,
  345,  346,  347,  348,  349,  350,  351,  352,  353,  354,
  355,  356,  357,  358,  359,  360,  361,  362,  363,  364,
  365,  366,  367,  368,  369,  370,  371,  372,  373,  374,
   -1,   -1,  265,  266,  267,   -1,  269,   -1,  271,   -1,
   -1,   -1,   -1,  388,   -1,   -1,   -1,   -1,   -1,  394,
  339,  340,  341,  342,  343,  344,  345,  346,  347,  348,
  349,  350,  351,  352,  353,  354,  355,  356,  357,  358,
  359,  360,  361,  362,  363,  364,  365,  366,  367,  368,
  369,  370,  371,  372,  373,  374,   -1,   -1,  265,  266,
  267,   -1,  269,   -1,  271,   -1,   -1,   -1,   -1,  388,
   -1,   -1,   -1,   -1,   -1,  394,  339,  340,  341,  342,
  343,  344,  345,  346,  347,  348,  349,  350,  351,  352,
  353,  354,  355,  356,  357,  358,  359,  360,  361,  362,
  363,  364,  365,  366,  367,  368,  369,  370,  371,  372,
  373,  374,   -1,   -1,  265,  266,  267,   -1,  269,   -1,
  271,   -1,   -1,   -1,   -1,  388,   -1,   -1,   -1,   -1,
   -1,  394,  339,  340,  341,  342,  343,  344,  345,  346,
  347,  348,  349,  350,  351,  352,  353,  354,  355,  356,
  357,  358,  359,  360,  361,  362,  363,  364,  365,  366,
  367,  368,  369,  370,  371,  372,  373,  374,   -1,   -1,
  265,  266,   -1,   -1,  269,   -1,  271,   -1,   -1,   -1,
   -1,  388,   -1,   -1,   -1,   -1,   -1,  394,  339,  340,
  341,  342,  343,  344,  345,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,  356,  357,  358,  359,  360,
  361,  362,  363,  364,  365,  366,  367,  368,  369,  370,
  371,  372,  373,  374,   -1,   -1,  265,  266,   -1,   -1,
  269,   -1,  271,   -1,   -1,   -1,   -1,  388,   -1,   -1,
   -1,   -1,   -1,  394,  339,  340,  341,  342,  343,  344,
  345,  346,  347,  348,  349,  350,  351,  352,  353,  354,
  355,  356,  357,  358,  359,  360,  361,  362,  363,  364,
  365,  366,  367,  368,  369,  370,  371,   -1,   -1,   -1,
   -1,   -1,  265,  266,   -1,   -1,  269,   -1,  271,   -1,
   -1,   -1,   -1,  388,   -1,   -1,   -1,   -1,   -1,  394,
  339,  340,  341,  342,  343,  344,  345,  346,  347,  348,
  349,  350,  351,  352,  353,  354,  355,  356,  357,  358,
  359,  360,  361,  362,  363,  364,  365,  366,  367,  368,
  369,  370,  371,   -1,   -1,   -1,   -1,   -1,  265,  266,
   -1,   -1,  269,   -1,  271,   -1,   -1,   -1,   -1,  388,
   -1,   -1,   -1,   -1,   -1,  394,  339,  340,  341,  342,
  343,  344,  345,  346,  347,  348,  349,  350,  351,  352,
  353,  354,  355,  356,  357,  358,  359,  360,  361,  362,
  363,  364,  365,  366,  367,  368,  369,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  265,  266,   -1,   -1,  269,   -1,
  271,   -1,   -1,   -1,   -1,  388,   -1,   -1,   -1,   -1,
   -1,  394,  339,  340,  341,  342,  343,  344,  345,  346,
  347,  348,  349,  350,  351,  352,  353,  354,  355,  356,
  357,  358,  359,  360,  361,  362,  363,  364,  365,  366,
  367,  368,  369,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  265,  266,   -1,   -1,  269,   -1,  271,   -1,   -1,   -1,
   -1,  388,   -1,   -1,   -1,   -1,   -1,  394,  339,  340,
  341,  342,  343,  344,  345,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,  356,  357,  358,  359,  360,
  361,  362,  363,  364,  365,  366,  367,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  265,  266,   -1,   -1,
  269,   -1,  271,   -1,   -1,   -1,   -1,  388,   -1,   -1,
   -1,   -1,   -1,  394,  339,  340,  341,  342,  343,  344,
  345,  346,  347,  348,  349,  350,  351,  352,  353,  354,
  355,  356,  357,  358,  359,  360,  361,  362,  363,  364,
  365,  366,  367,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  265,  266,   -1,   -1,  269,   -1,  271,   -1,
   -1,   -1,   -1,  388,   -1,   -1,   -1,   -1,   -1,  394,
  339,  340,  341,  342,  343,  344,  345,  346,  347,  348,
  349,  350,  351,  352,  353,  354,  355,  356,  357,  358,
  359,  360,  361,  362,  363,  364,  365,  366,  367,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  265,  266,
   -1,   -1,  269,   -1,  271,   -1,   -1,   -1,   -1,  388,
   -1,   -1,   -1,   -1,   -1,  394,  339,  340,  341,  342,
  343,  344,  345,  346,  347,  348,  349,  350,  351,  352,
  353,  354,  355,  356,  357,  358,  359,  360,  361,  362,
  363,  364,  365,  366,  367,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  265,  266,   -1,   -1,  269,   -1,
  271,   -1,   -1,   -1,   -1,  388,   -1,   -1,   -1,   -1,
   -1,  394,  339,  340,  341,  342,  343,  344,  345,  346,
  347,  348,  349,  350,  351,  352,  353,  354,  355,  356,
  357,  358,  359,  360,  361,  362,  363,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  265,  266,   -1,   -1,  269,   -1,  271,   -1,   -1,   -1,
   -1,  388,   -1,   -1,   -1,   -1,   -1,  394,  339,  340,
  341,  342,  343,  344,  345,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,  356,  357,  358,  359,  360,
  361,  362,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  265,  266,   -1,   -1,
  269,   -1,  271,   -1,   -1,   -1,   -1,  388,   -1,   -1,
   -1,   -1,   -1,  394,  339,  340,  341,  342,  343,  344,
  345,  346,  347,  348,  349,  350,  351,  352,  353,  354,
  355,  356,  357,  358,  359,  360,  361,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  266,  267,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  388,   -1,   -1,   -1,   -1,   -1,  394,
  339,  340,  341,  342,  343,  344,  345,  346,  347,  348,
  349,  350,  351,  352,  353,  354,  355,  356,  357,  358,
  359,  360,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  267,   -1,  269,   -1,  271,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  388,
   -1,   -1,   -1,   -1,   -1,  394,  342,  343,  344,  345,
  346,  347,  348,  349,  350,  351,  352,  353,  354,  355,
   -1,  357,  358,  359,  360,  361,  362,  363,  364,  365,
  366,  367,  368,  369,  370,  371,  372,  373,  374,  267,
   -1,   -1,  378,  271,   -1,  381,  382,   -1,  384,   -1,
   -1,  387,   -1,  389,  390,  391,  339,  393,  341,   -1,
  343,  344,  345,  346,  347,  348,  349,  350,  351,  352,
  353,  354,  355,   -1,  357,  358,  359,  360,  361,  362,
  363,  364,  365,  366,  367,  368,  369,  370,  371,  372,
  373,  374,  267,   -1,   -1,  378,   -1,   -1,  381,  382,
   -1,  384,   -1,   -1,  387,   -1,  389,  390,  391,   -1,
  393,   -1,   -1,  341,  342,  343,  344,  345,  346,  347,
  348,  349,  350,  351,  352,  353,  354,  355,   -1,  357,
  358,  359,  360,  361,  362,  363,  364,  365,  366,  367,
  368,  369,  370,  371,  372,  373,  374,  267,   -1,   -1,
  378,   -1,   -1,  381,  382,   -1,  384,   -1,   -1,  387,
   -1,  389,  390,  391,   -1,  393,  341,  342,  343,  344,
  345,  346,  347,  348,  349,  350,  351,  352,  353,  354,
  355,   -1,  357,  358,  359,  360,  361,  362,  363,  364,
  365,  366,  367,  368,  369,  370,  371,  372,  373,  374,
  267,   -1,   -1,  378,   -1,   -1,  381,  382,   -1,  384,
   -1,   -1,  387,   -1,  389,  390,  391,   -1,  393,   -1,
   -1,  341,  342,  343,  344,  345,  346,  347,  348,  349,
  350,  351,  352,  353,  354,  355,   -1,  357,  358,  359,
  360,  361,  362,  363,  364,  365,  366,  367,  368,  369,
  370,  371,  372,  373,  374,  267,   -1,   -1,  378,   -1,
   -1,  381,  382,   -1,  384,   -1,   -1,  387,   -1,  389,
  390,  391,   -1,  393,   -1,  342,  343,  344,  345,  346,
  347,  348,  349,  350,  351,  352,  353,  354,  355,  356,
  357,  358,  359,  360,  361,  362,  363,  364,  365,  366,
  367,  368,  369,  370,  371,  372,  373,  374,  267,   -1,
   -1,  378,   -1,   -1,  381,  382,   -1,  384,   -1,   -1,
  387,   -1,  389,  390,  391,   -1,  393,  339,   -1,   -1,
  342,  343,  344,  345,  346,  347,  348,  349,  350,  351,
  352,  353,  354,  355,   -1,  357,  358,  359,  360,  361,
  362,  363,  364,  365,  366,  367,  368,  369,  370,  371,
  372,  373,  374,  267,   -1,   -1,  378,   -1,   -1,  381,
  382,   -1,  384,   -1,   -1,  387,   -1,  389,  390,  391,
   -1,  393,  341,  342,  343,  344,  345,  346,  347,  348,
  349,  350,  351,  352,  353,  354,  355,   -1,  357,  358,
  359,  360,  361,  362,  363,  364,  365,  366,  367,  368,
  369,  370,  371,  372,  373,  374,  267,   -1,   -1,  378,
   -1,   -1,  381,  382,   -1,  384,   -1,   -1,  387,   -1,
  389,  390,  391,   -1,  393,   -1,   -1,   -1,  342,  343,
  344,  345,  346,  347,  348,  349,  350,  351,  352,  353,
  354,  355,  267,  357,  358,  359,  360,  361,  362,  363,
  364,  365,  366,  367,  368,  369,  370,  371,  372,  373,
  374,   -1,   -1,   -1,  378,   -1,   -1,  381,  382,   -1,
  384,   -1,   -1,  387,   -1,  389,  390,  391,  267,  393,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,
  361,  362,  363,  364,  365,  366,  367,  368,  369,  370,
  371,  372,  373,  374,   -1,   -1,   -1,  378,   -1,   -1,
  381,  382,   -1,  384,   -1,   -1,  387,   -1,  389,  390,
  391,   -1,  393,   -1,   -1,  360,  361,  362,  363,  364,
  365,  366,  367,  368,  369,  370,  371,  372,  373,  374,
   -1,   -1,   -1,  378,   -1,   -1,  381,  382,   -1,  384,
   -1,   -1,  387,   -1,  389,  390,  391,   -1,  393,   -1,
   -1,   -1,  361,  362,  363,  364,  365,  366,  367,  368,
  369,  370,  371,  372,  373,  374,   -1,   -1,   -1,  378,
   -1,   -1,  381,  382,   -1,  384,   -1,   -1,  387,   -1,
  389,  390,  391,   -1,  393,
};
#define YYFINAL 1
#ifndef YYDEBUG
#define YYDEBUG 0
#endif
#define YYMAXTOKEN 396
#if YYDEBUG
char *yyname[] = {
"end-of-file",0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,"VAR","EXPR","ARRAY","STRUCT",
"UNION","ENUM","COMP","HASH","NL","SEMI","MOD","OC","CC","DOLLAR","DOTDOTDOT",
"ENDFILE","GLOBAL","AUTO","STATIC","CONST","POLY","INTEGER","NATURAL",
"RATIONAL","REAL","STRING","FILET","MUTEX","SEMAPHORE","CONTINUATION","THREAD",
"VOID","BOOL","FUNCTION","FUNC","EXCEPTION","RAISE","TYPEDEF","IMPORT","NEW",
"ANONINIT","NAMESPACE","PUBLIC","PROTECTED","EXTEND","IF","ELSE","WHILE","DO",
"FOR","SWITCH","BREAK","CONTINUE","RETURNTOK","FORK","CASE","DEFAULT","TRY",
"CATCH","TWIXT","NAME","TYPENAME","NAMESPACENAME","COMMAND","NAMECOMMAND",
"TEN_NUM","OCTAL0_NUM","OCTAL_NUM","BINARY_NUM","HEX_NUM","TEN_FLOAT",
"OCTAL0_FLOAT","OCTAL_FLOAT","BINARY_FLOAT","HEX_FLOAT","CHAR_CONST",
"STRING_CONST","POLY_CONST","THREAD_CONST","COMMENT_CONST","VOIDVAL","BOOLVAL",
"DARROW","POUND","COMMA","ASSIGN","ASSIGNPLUS","ASSIGNMINUS","ASSIGNTIMES",
"ASSIGNDIVIDE","ASSIGNDIV","ASSIGNMOD","ASSIGNPOW","ASSIGNSHIFTL",
"ASSIGNSHIFTR","ASSIGNLXOR","ASSIGNLAND","ASSIGNLOR","QUEST","COLON","OR","AND",
"LOR","LXOR","LAND","EQ","NE","LT","GT","LE","GE","SHIFTL","SHIFTR","PLUS",
"MINUS","TIMES","DIVIDE","DIV","POW","UNIONCAST","UMINUS","BANG","FACT","LNOT",
"INC","DEC","STAR","STARSTAR","AMPER","THREADID","OS","CS","DOT","ARROW",
"STAROS","CALL","OP","CP","POINTER","COLONCOLON",
};
char *yyrule[] = {
"$accept : lines",
"lines : lines pcommand",
"lines :",
"pcommand : command",
"pcommand : error reset eend",
"eend : NL",
"eend : SEMI",
"eend : ENDFILE",
"ignorenl :",
"attendnl :",
"reset :",
"not_command :",
"opt_nl : NL",
"opt_nl :",
"opt_semi : SEMI",
"opt_semi :",
"opt_comma : COMMA",
"opt_comma :",
"command : not_command expr reset NL",
"command : not_command expr POUND expr reset NL",
"command : not_command statement reset opt_nl",
"command : not_command COMMAND opt_exprs reset opt_semi NL",
"command : not_command NAMECOMMAND opt_rawnames reset opt_semi NL",
"command : NL",
"command : ENDFILE",
"opt_rawnames : rawnames",
"opt_rawnames :",
"rawnames : rawname COMMA rawnames",
"rawnames : rawname",
"rawname : rawnamespace rawatom",
"rawatom : NAME",
"rawatom : NAMESPACENAME",
"rawatom : TYPENAME",
"rawnamespace : rawnamespace rawatom COLONCOLON",
"rawnamespace :",
"fulltype : namespace TYPENAME",
"fulltype : TYPENAME",
"fullname : namespace namespacename",
"fullname : namespacename",
"namespace : namespace NAMESPACENAME COLONCOLON",
"namespace : NAMESPACENAME COLONCOLON",
"namespacename : NAME",
"namespacename : NAMESPACENAME",
"namespace_start :",
"namespace_end :",
"block : block_start statements block_end",
"block_start : OC namespace_start",
"block_end : namespace_end CC",
"statements : statement statements",
"statements :",
"statement : IF ignorenl namespace_start OP expr CP statement namespace_end attendnl",
"statement : IF ignorenl namespace_start OP expr CP statement ELSE statement namespace_end attendnl",
"statement : WHILE ignorenl namespace_start OP expr CP statement namespace_end attendnl",
"statement : DO ignorenl namespace_start statement WHILE OP expr CP namespace_end attendnl",
"statement : FOR ignorenl namespace_start OP for_exprs CP statement namespace_end attendnl",
"statement : SWITCH ignorenl namespace_start OP expr CP case_block namespace_end attendnl",
"statement : UNION SWITCH ignorenl namespace_start OP expr CP union_case_block namespace_end attendnl",
"statement : BREAK SEMI",
"statement : CONTINUE SEMI",
"statement : RETURNTOK opt_expr SEMI",
"statement : expr SEMI",
"statement : block",
"statement : SEMI",
"statement : func_decl doc_string opt_func_body namespace_end",
"statement : opt_publish EXCEPTION ignorenl NAME namespace_start opt_argdecls namespace_end doc_string SEMI attendnl",
"statement : RAISE fullname OP opt_exprs CP SEMI",
"statement : opt_publish TYPEDEF ignorenl typenames SEMI attendnl",
"statement : opt_publish TYPEDEF ignorenl type typenames SEMI attendnl",
"$$1 :",
"statement : publish_extend NAMESPACE ignorenl namespacename $$1 OC statements CC attendnl",
"statement : opt_publish IMPORT ignorenl fullname SEMI attendnl",
"statement : TRY ignorenl statement catches attendnl",
"statement : TWIXT ignorenl namespace_start OP opt_expr SEMI opt_expr CP statement namespace_end attendnl",
"for_exprs : opt_expr SEMI for_exprs",
"for_exprs : opt_expr",
"catches : catches catch",
"catches :",
"catch : CATCH fullname namespace_start args doc_string block namespace_end",
"opt_func_body : func_body",
"opt_func_body : SEMI",
"$$2 :",
"func_body : $$2 block_or_expr",
"block_or_expr : block",
"block_or_expr : attendnl ASSIGN simpleexpr SEMI",
"see_comment :",
"doc_string : see_comment opt_comment",
"opt_comment : COMMENT_CONST",
"opt_comment :",
"case_block : block_start cases block_end",
"cases : case cases",
"cases :",
"case : CASE expr COLON statements",
"case : DEFAULT COLON statements",
"union_case_block : block_start union_cases block_end",
"union_cases : union_case union_cases",
"union_cases :",
"union_case : CASE namespace_start NAME opt_name COLON statements namespace_end",
"union_case : DEFAULT COLON statements",
"opt_name : NAME",
"opt_name :",
"atoms : NAME COMMA atoms",
"atoms : NAME",
"typenames : typename COMMA typenames",
"typenames : typename",
"typename : TYPENAME",
"typename : NAME",
"initnames : name opt_init next_decl initnames",
"initnames : name opt_init",
"name : NAME",
"next_decl : COMMA",
"func_decl : func_name namespace_start opt_argdefines CP",
"func_name : decl NAME OP",
"func_name : decl FUNCTION NAME OP",
"func_name : FUNCTION ignorenl NAME OP",
"opt_init : ASSIGN simpleexpr",
"opt_init : ASSIGN init",
"opt_init :",
"decl : publish class type opt_nl",
"decl : class type opt_nl",
"decl : publish type opt_nl",
"decl : type opt_nl",
"decl : publish class opt_nl",
"decl : class opt_nl",
"decl : publish opt_nl",
"subscripts : opt_argdecls subscripts",
"subscripts : OS opt_stars CS subscripts",
"subscripts : OS dotdotdots CS subscripts",
"subscripts : OS dims CS subscripts",
"subscripts : OS type CS subscripts",
"subscripts :",
"type : subtype subscripts",
"type : TIMES type",
"type : STARSTAR type",
"type : AND type",
"type : LAND type",
"subtype : basetype",
"subtype : STRUCT OC struct_members CC",
"subtype : UNION OC union_members CC",
"subtype : ENUM OC atoms CC",
"subtype : OP type CP",
"subtype : fulltype",
"opt_type : type",
"opt_type :",
"basetype : POLY",
"basetype : INTEGER",
"basetype : RATIONAL",
"basetype : REAL",
"basetype : STRING",
"basetype : FILET",
"basetype : MUTEX",
"basetype : SEMAPHORE",
"basetype : CONTINUATION",
"basetype : THREAD",
"basetype : VOID",
"basetype : BOOL",
"opt_stars : stars",
"opt_stars :",
"stars : stars COMMA TIMES",
"stars : TIMES",
"dotdotdots : dotdotdots COMMA DOTDOTDOT",
"dotdotdots : DOTDOTDOT",
"dims : simpleexpr COMMA dims",
"dims : simpleexpr",
"struct_members : opt_type atoms SEMI struct_members",
"struct_members :",
"union_members : opt_type atoms SEMI union_members",
"union_members :",
"class : GLOBAL",
"class : AUTO",
"class : STATIC",
"class : CONST",
"opt_publish : publish",
"opt_publish :",
"publish : PUBLIC",
"publish : PROTECTED",
"publish_extend : opt_publish",
"publish_extend : EXTEND",
"opt_argdecls : OP argdecls CP",
"opt_argdecls : OP CP",
"argdecls : argdecl COMMA argdecls",
"argdecls : argdecl opt_dotdotdot",
"argdecl : type NAME",
"argdecl : type",
"args : OP opt_argdefines CP",
"opt_argdefines : ignorenl argdefines",
"opt_argdefines : ignorenl",
"argdefines : argdefine COMMA argdefines",
"argdefines : argdefine opt_dotdotdot",
"argdefine : opt_type NAME",
"argdefine : type",
"opt_dotdotdot : DOTDOTDOT",
"opt_dotdotdot :",
"opt_expr : expr",
"opt_expr :",
"expr : comma_expr",
"expr : decl initnames",
"comma_expr : simpleexpr",
"comma_expr : comma_expr COMMA simpleexpr",
"opt_exprs : exprs",
"opt_exprs :",
"exprs : simpleexpr COMMA exprs",
"exprs : simpleexpr",
"opt_actuals : actuals",
"opt_actuals :",
"actuals : simpleexpr COMMA actuals",
"actuals : simpleexpr opt_dotdotdot",
"func_right : attendnl ASSIGN simpleexpr",
"$$3 :",
"func_right : $$3 block",
"simpleexpr : simpleexpr assignop simpleexpr",
"simpleexpr : opt_type FUNC namespace_start args doc_string func_right namespace_end",
"simpleexpr : MOD integer",
"simpleexpr : TIMES simpleexpr",
"simpleexpr : STARSTAR simpleexpr",
"simpleexpr : LAND simpleexpr",
"simpleexpr : AND simpleexpr",
"simpleexpr : MINUS simpleexpr",
"simpleexpr : LNOT simpleexpr",
"simpleexpr : BANG simpleexpr",
"simpleexpr : simpleexpr BANG",
"simpleexpr : INC simpleexpr",
"simpleexpr : simpleexpr INC",
"simpleexpr : DEC simpleexpr",
"simpleexpr : simpleexpr DEC",
"simpleexpr : FORK simpleexpr",
"simpleexpr : simpleexpr PLUS simpleexpr",
"simpleexpr : simpleexpr MINUS simpleexpr",
"simpleexpr : simpleexpr TIMES simpleexpr",
"simpleexpr : simpleexpr DIVIDE simpleexpr",
"simpleexpr : simpleexpr DIV simpleexpr",
"simpleexpr : simpleexpr MOD simpleexpr",
"simpleexpr : simpleexpr STARSTAR simpleexpr",
"simpleexpr : simpleexpr SHIFTL simpleexpr",
"simpleexpr : simpleexpr SHIFTR simpleexpr",
"simpleexpr : simpleexpr QUEST simpleexpr COLON simpleexpr",
"simpleexpr : simpleexpr LXOR simpleexpr",
"simpleexpr : simpleexpr LAND simpleexpr",
"simpleexpr : simpleexpr LOR simpleexpr",
"simpleexpr : simpleexpr AND simpleexpr",
"simpleexpr : simpleexpr OR simpleexpr",
"simpleexpr : simpleexpr EQ simpleexpr",
"simpleexpr : simpleexpr NE simpleexpr",
"simpleexpr : simpleexpr LT simpleexpr",
"simpleexpr : simpleexpr GT simpleexpr",
"simpleexpr : simpleexpr LE simpleexpr",
"simpleexpr : simpleexpr GE simpleexpr",
"simpleexpr : fullname",
"simpleexpr : TEN_NUM",
"simpleexpr : OCTAL_NUM",
"simpleexpr : OCTAL0_NUM",
"simpleexpr : BINARY_NUM",
"simpleexpr : HEX_NUM",
"simpleexpr : TEN_FLOAT",
"simpleexpr : OCTAL_FLOAT",
"simpleexpr : OCTAL0_FLOAT",
"simpleexpr : BINARY_FLOAT",
"simpleexpr : HEX_FLOAT",
"simpleexpr : CHAR_CONST",
"simpleexpr : STRING_CONST",
"simpleexpr : VOIDVAL",
"simpleexpr : BOOLVAL",
"simpleexpr : OS CS",
"simpleexpr : OP type CP namespace_start init namespace_end",
"simpleexpr : OP OS stars CS CP namespace_start arrayinit namespace_end",
"simpleexpr : OP OS dims CS CP namespace_start opt_arrayinit namespace_end",
"simpleexpr : OP OS type CS CP namespace_start opt_hashinit namespace_end",
"simpleexpr : type DOT NAME",
"simpleexpr : OP type DOT NAME CP simpleexpr",
"simpleexpr : DOLLAR opt_integer",
"simpleexpr : DOT",
"simpleexpr : OP expr CP",
"simpleexpr : OP block CP",
"simpleexpr : simpleexpr STAROS dims CS",
"simpleexpr : simpleexpr OS dims CS",
"simpleexpr : simpleexpr OP opt_actuals CP",
"simpleexpr : simpleexpr DOT NAME",
"simpleexpr : simpleexpr ARROW NAME",
"opt_integer : integer",
"opt_integer :",
"assignop : ASSIGNPLUS",
"assignop : ASSIGNMINUS",
"assignop : ASSIGNTIMES",
"assignop : ASSIGNDIVIDE",
"assignop : ASSIGNDIV",
"assignop : ASSIGNMOD",
"assignop : ASSIGNPOW",
"assignop : ASSIGNSHIFTL",
"assignop : ASSIGNSHIFTR",
"assignop : ASSIGNLXOR",
"assignop : ASSIGNLAND",
"assignop : ASSIGNLOR",
"assignop : ASSIGN",
"integer : TEN_NUM",
"integer : OCTAL_NUM",
"integer : OCTAL0_NUM",
"integer : BINARY_NUM",
"integer : HEX_NUM",
"opt_arrayinit : arrayinit",
"opt_arrayinit : OC CC",
"opt_arrayinit :",
"arrayinit : OC arrayelts opt_comma opt_dotdotdot CC",
"arrayinit : OC OS namespace_start compnames comparray CS comprehension namespace_end CC",
"comprehension : ASSIGN arrayelt",
"comprehension : block",
"compnames : compname COMMA compnames",
"compnames : compname",
"compname : NAME",
"comparray :",
"arrayelts : arrayelts COMMA arrayelt",
"arrayelts : arrayelt",
"arrayelt : simpleexpr",
"arrayelt : init",
"opt_hashinit : hashinit",
"opt_hashinit : OC CC",
"opt_hashinit :",
"hashinit : OC hashelts opt_comma CC",
"hashelts : hashelts COMMA hashelt",
"hashelts : hashelt",
"hashelt : simpleexpr DARROW hashvalue",
"hashelt : DARROW hashvalue",
"hashvalue : simpleexpr",
"hashvalue : init",
"structinit : OC structelts opt_comma CC",
"structelts : structelts COMMA structelt",
"structelts : structelt",
"structelt : NAME ASSIGN simpleexpr",
"structelt : NAME ASSIGN init",
"init : arrayinit",
"init : structinit",
"init : hashinit",
"init : OC CC",
};
#endif
#ifdef YYSTACKSIZE
#undef YYMAXDEPTH
#define YYMAXDEPTH YYSTACKSIZE
#else
#ifdef YYMAXDEPTH
#define YYSTACKSIZE YYMAXDEPTH
#else
#define YYSTACKSIZE 500
#define YYMAXDEPTH 500
#endif
#endif
int yydebug;
int yynerrs;
int yyerrflag;
int yychar;
short *yyssp;
YYSTYPE *yyvsp;
YYSTYPE yyval;
YYSTYPE yylval;
short yyss[YYSTACKSIZE];
YYSTYPE yyvs[YYSTACKSIZE];
#define yystacksize YYSTACKSIZE
#line 1560 "gram.y"

static void
DeclListMark (void *object)
{
    DeclListPtr	dl = object;

    MemReference (dl->next);
    MemReference (dl->init);
}

DataType DeclListType = { DeclListMark, 0, "DeclListType" };

DeclListPtr
NewDeclList (Atom name, ExprPtr init, DeclListPtr next)
{
    ENTER ();
    DeclListPtr	dl;

    dl = ALLOCATE (&DeclListType, sizeof (DeclList));
    dl->next = next;
    dl->name = name;
    dl->symbol = 0;
    dl->init = init;
    RETURN (dl);
}

static void
MemListMark (void *object)
{
    MemListPtr	ml = object;

    MemReference (ml->next);
    MemReference (ml->type);
    MemReference (ml->atoms);
}

DataType MemListType = { MemListMark, 0, "MemListType" };

MemListPtr
NewMemList (AtomListPtr atoms, Type *type, MemListPtr next)
{
    ENTER ();
    MemListPtr	ml;

    ml = ALLOCATE (&MemListType, sizeof (MemList));
    ml->next = next;
    ml->type = type;
    ml->atoms = atoms;
    RETURN (ml);
}


extern NamespacePtr	CurrentNamespace;
FramePtr	CurrentFrame;

Value
lookupVar (char *ns, char *n)
{
    ENTER ();
    Value	    v;
    SymbolPtr	    symbol;
    NamespacePtr    namespace;
    Bool	    search;

    search = True;
    namespace = CurrentNamespace;
    if (ns)
    {
	symbol = NamespaceFindName (CurrentNamespace, AtomId (ns), True);
	if (symbol && symbol->symbol.class == class_namespace)
	    namespace = symbol->namespace.namespace;
	else
	    namespace = 0;
    }
    if (namespace)
	symbol = NamespaceFindName (namespace, AtomId (n), search);
    else
	symbol = 0;
    if (symbol && symbol->symbol.class == class_global)
	v = BoxValue (symbol->global.value, 0);
    else
	v = Void;
    RETURN (v);
}

void
setVar (NamespacePtr namespace, char *n, Value v, Type *type)
{
    ENTER ();
    Atom	atom;
    SymbolPtr	symbol;

    atom = AtomId (n);
    symbol = NamespaceFindName (namespace, atom, True);
    if (!symbol)
	symbol = NamespaceAddName (namespace,
				   NewSymbolGlobal (atom, type),
				   publish_private);
    if (symbol->symbol.class == class_global)
	BoxValueSet (symbol->global.value, 0, v);
    EXIT ();
}

void
GetNamespace (NamespacePtr *scope, FramePtr *fp, CodePtr *cp)
{
    *scope = CurrentNamespace;
    *fp = CurrentFrame;
    if (CurrentFrame)
	*cp = CurrentFrame->function->func.code;
    else
	*cp = 0;
}

ExprPtr
BuildName (char *ns, char *n)
{
    ENTER ();
    SymbolPtr	    symbol, ns_symbol = 0;
    Atom	    atom, ns_atom = 0;
    Bool	    search = True;
    NamespacePtr    namespace = CurrentNamespace;
    ExprPtr	    e;
    Bool	    ns_privateFound = False;
    Bool	    privateFound = False;

    if (ns)
    {
	ns_atom = AtomId (ns);
	ns_symbol = NamespaceFindName (namespace, ns_atom, search);
	if (ns_symbol && ns_symbol->symbol.class == class_namespace)
	    namespace = ns_symbol->namespace.namespace;
	else
	{
	    if (!ns_symbol)
		ns_privateFound = NamespaceIsNamePrivate (namespace, ns_atom, search);
	    namespace = 0;
	}
	search = False;
    }
    atom = AtomId (n);
    if (namespace)
    {
	symbol = NamespaceFindName (namespace, atom, search);
	if (!symbol)
	    privateFound = NamespaceIsNamePrivate (namespace, atom, search);
    }
    else
	symbol = 0;
    e = NewExprAtom (atom, symbol, privateFound);
    if (ns_atom)
	e = NewExprTree (COLONCOLON, NewExprAtom (ns_atom, ns_symbol, ns_privateFound), e);
    RETURN (e);
}

static Value
AtomString (Atom id)
{
    ENTER ();
    RETURN (NewStrString (AtomName (id)));
}

ExprPtr
BuildRawname (ExprPtr colonnames, Atom name)
{
    ENTER ();
    int     len;
    Value   array;
    ExprPtr e;

    len = 1;
    for (e = colonnames; e; e = e->tree.left)
	len++;
    array = NewArray (False, False, typePrim[rep_string], 1, &len);
    len--;
    ArrayValueSet (&array->array, len, AtomString (name));
    e = colonnames;
    while (--len >= 0)
    {
	ArrayValueSet (&array->array, len, AtomString (e->tree.right->atom.atom));
	e = e->tree.left;
    }
    e = NewExprConst (POLY_CONST, array);
    RETURN (e);
}

	    
ExprPtr
BuildFullname (ExprPtr left, Atom atom)
{
    ENTER ();
    NamespacePtr    namespace;
    SymbolPtr	    symbol;
    Bool	    search;
    ExprPtr	    nameExpr;
    Bool	    privateFound = False;

    if (left)
    {
	if (left->base.tag == COLONCOLON)
	    symbol = left->tree.right->atom.symbol;
	else
	    symbol = left->atom.symbol;
	if (symbol && symbol->symbol.class == class_namespace)
	    namespace = symbol->namespace.namespace;
	else
	    namespace = 0;
        search = False;
    }
    else
    {
	namespace = CurrentNamespace;
	search = True;
    }
    if (namespace)
    {
	symbol = NamespaceFindName (namespace, atom, search);
	if (!symbol)
	    privateFound = NamespaceIsNamePrivate (namespace, atom, search);
    }
    else
	symbol = 0;
    nameExpr = NewExprAtom (atom, symbol, privateFound);
    if (left)
	nameExpr = NewExprTree (COLONCOLON, left, nameExpr);
    RETURN (nameExpr);
}

ExprPtr
BuildCall (char *scope, char *name, int nargs, ...)
{
    ENTER ();
    va_list	    alist;
    ExprPtr	    args, *prev;
    ExprPtr	    f;
    ExprPtr	    e;

    va_start (alist, nargs);
    prev = &args;
    args = 0;
    while (nargs--)
    {
	*prev = NewExprTree (COMMA, va_arg (alist, ExprPtr), 0);
	prev = &(*prev)->tree.right;
    }
    va_end (alist);
    f = BuildName (scope, name);
    e = NewExprTree (OP, f, args);
#ifdef DEBUG
    printExpr (stdout, e, -1, 0);
    printf ("\n");
#endif
    RETURN (e);
}

/*
 * Walk for() loop arguments and normalize the list
 */
static Expr *
ParseCanonFor (Expr *expr)
{
    if (!expr || !expr->tree.right) {
        ParseError ("Too few exprs in for()\n");
        return 0;
    }
    if (!expr->tree.right->tree.right) {
        /* 2-argument for() */
        expr = NewExprTree(FOR, 0, expr);
    }
    if (expr->tree.right->tree.right->tree.right) {
        ParseError ("Too many exprs in for()\n");
        return 0;
    }
    return expr;
}

/*
 * Walk a type structure and resolve any type names
 */
static CanonTypeResult
ParseCanonType (TypePtr type, Bool forwardAllowed)
{
    ArgType	    *arg;
    int		    n;
    CanonTypeResult ret = CanonTypeDefined, t;
    Bool	    anyResolved;
    
    if (!type)
    {
	ParseError ("Type missing inside compiler");
	return False;
    }
    switch (type->base.tag) {
    case type_prim:
	break;
    case type_name:
	if (!TypeNameType(type))
	{
	    if (!type->name.name)
	    {
		ExprPtr e;
		e = type->name.expr;
		if (e->base.tag == COLONCOLON)
		    e = e->tree.right;
		type->name.name = e->atom.symbol;
		if (!type->name.name)
		{
		    ParseError ("No typedef \"%A\" in namespace",
				e->atom.atom);
		    ret = CanonTypeUndefined;
		    break;
		}
	    }
	    if (type->name.name->symbol.class != class_typedef)
	    {
		ParseError ("Symbol \"%A\" not a typedef", 
			    type->name.name->symbol.name);
		ret = CanonTypeUndefined;
	    }
	    else if (!TypeNameType(type))
	    {
		if (!forwardAllowed)
		    ParseError ("Typedef \"%A\" not defined yet", 
				type->name.name->symbol.name);
		ret = CanonTypeForward;
	    }
	    else
	    {
		ret = ParseCanonType (TypeNameType(type), forwardAllowed);
	    }
	}
	break;
    case type_ref:
	ret = ParseCanonType (type->ref.ref, True);
	if (ret == CanonTypeForward)
	    ret = CanonTypeDefined;
	break;
    case type_func:
	if (type->func.ret)
	    ret = ParseCanonType (type->func.ret, forwardAllowed);
	for (arg = type->func.args; arg; arg = arg->next)
	{
	    t = ParseCanonType (arg->type, forwardAllowed);
	    if (t < ret)
		ret = t;
	}
	break;
    case type_array:
	ret = ParseCanonType (type->array.type, forwardAllowed);
	break;
    case type_hash:
	ret = ParseCanonType (type->hash.type, forwardAllowed);
	t = ParseCanonType (type->hash.keyType, forwardAllowed);
	if (t < ret)
	    ret = t;
	break;
    case type_struct:
	for (n = 0; n < type->structs.structs->nelements; n++)
	{
	    StructType   *st = type->structs.structs;

	    t = ParseCanonType (BoxTypesElements(st->types)[n], forwardAllowed);
	    if (t < ret)
		ret = t;
	}
	break;
    case type_union:
	anyResolved = False;
	for (n = 0; n < type->structs.structs->nelements; n++)
	{
	    StructType   *st = type->structs.structs;

	    t = ParseCanonType (BoxTypesElements(st->types)[n], True);
	    if (t < ret)
		ret = t;
	    if (t == CanonTypeDefined)
		anyResolved = True;
	}
	if (ret == CanonTypeForward)
	{
	    if (anyResolved)
		ret = CanonTypeDefined;
	    else if (!forwardAllowed)
		ParseError ("No member of '%T' defined yet", type);
	}
	break;
    case type_types:
	break;
    }
    return ret;
}

static SymbolPtr
ParseNewSymbol (Publish publish, Class class, Type *type, Atom name)
{
    ENTER ();
    SymbolPtr	s = 0;
    
    if (class == class_undef)
	class = funcDepth ? class_auto : class_global;

    if (class == class_namespace || 
	(class == class_typedef && type == 0) || 	
	ParseCanonType (type, False) == CanonTypeDefined)
    {
	switch (class) {
	case class_const:
	    s = NewSymbolConst (name, type);
	    break;
	case class_global:
	    s = NewSymbolGlobal (name, type);
	    break;
	case class_static:
	    s = NewSymbolStatic (name, type);
	    break;
	case class_arg:
	    s = NewSymbolArg (name, type);
	    break;
	case class_auto:
	    s = NewSymbolAuto (name, type);
	    break;
	case class_exception:
	    s = NewSymbolException (name, type, Void);
	    break;
	case class_typedef:
	    /*
	     * Special case for typedefs --
	     * allow forward declaration of untyped
	     * typedef names, then hook the
	     * new type to the old name
	     */
	    if (type)
	    {
		s = NamespaceFindName (CurrentNamespace, name, False);
		if (s && s->symbol.class == class_typedef && !s->symbol.type)
		{
		    s->symbol.type = type;
		    RETURN (s);
		}
	    }
	    s = NewSymbolType (name, type);
	    break;
	case class_namespace:
	    s = NewSymbolNamespace (name, NewNamespace (CurrentNamespace));
	    break;
	case class_undef:
	    break;
	}
	if (s)
	    NamespaceAddName (CurrentNamespace, s, publish);
    }
    RETURN (s);
}

int
yywrap (void)
{
    if (LexInteractive())
	printf ("\n");
    if (CurrentFrame)
    {
	do_Debug_done ();
	return 0;
    }
    return 1;
}

extern	char *yytext;

void
ParseError (char *fmt, ...)
{
    va_list	args;

    if (LexFileName ())
	FilePrintf (FileStderr, "%A:%d: ",
		    LexFileName (), LexFileLine ());
    va_start (args, fmt);
    FileVPrintf (FileStderr, fmt, args);
    FilePrintf (FileStderr, "\n");
    va_end (args);
}

void
yyerror (char *msg)
{
    ParseError ("%s before %S", msg, yytext);
}
#line 2593 "gram.c"
#define YYABORT goto yyabort
#define YYREJECT goto yyabort
#define YYACCEPT goto yyaccept
#define YYERROR goto yyerrlab
int
yyparse()
{
    register int yym, yyn, yystate;
#if YYDEBUG
    register char *yys;
    extern char *getenv();

    if (yys = getenv("YYDEBUG"))
    {
        yyn = *yys;
        if (yyn >= '0' && yyn <= '9')
            yydebug = yyn - '0';
    }
#endif

    yynerrs = 0;
    yyerrflag = 0;
    yychar = (-1);

    yyssp = yyss;
    yyvsp = yyvs;
    *yyssp = yystate = 0;

yyloop:
    if (yyn = yydefred[yystate]) goto yyreduce;
    if (yychar < 0)
    {
        if ((yychar = yylex()) < 0) yychar = 0;
#if YYDEBUG
        if (yydebug)
        {
            yys = 0;
            if (yychar <= YYMAXTOKEN) yys = yyname[yychar];
            if (!yys) yys = "illegal-symbol";
            printf("%sdebug: state %d, reading %d (%s)\n",
                    YYPREFIX, yystate, yychar, yys);
        }
#endif
    }
    if ((yyn = yysindex[yystate]) && (yyn += yychar) >= 0 &&
            yyn <= YYTABLESIZE && yycheck[yyn] == yychar)
    {
#if YYDEBUG
        if (yydebug)
            printf("%sdebug: state %d, shifting to state %d\n",
                    YYPREFIX, yystate, yytable[yyn]);
#endif
        if (yyssp >= yyss + yystacksize - 1)
        {
            goto yyoverflow;
        }
        *++yyssp = yystate = yytable[yyn];
        *++yyvsp = yylval;
        yychar = (-1);
        if (yyerrflag > 0)  --yyerrflag;
        goto yyloop;
    }
    if ((yyn = yyrindex[yystate]) && (yyn += yychar) >= 0 &&
            yyn <= YYTABLESIZE && yycheck[yyn] == yychar)
    {
        yyn = yytable[yyn];
        goto yyreduce;
    }
    if (yyerrflag) goto yyinrecovery;
#ifdef lint
    goto yynewerror;
#endif
yynewerror:
    yyerror("syntax error");
#ifdef lint
    goto yyerrlab;
#endif
yyerrlab:
    ++yynerrs;
yyinrecovery:
    if (yyerrflag < 3)
    {
        yyerrflag = 3;
        for (;;)
        {
            if ((yyn = yysindex[*yyssp]) && (yyn += YYERRCODE) >= 0 &&
                    yyn <= YYTABLESIZE && yycheck[yyn] == YYERRCODE)
            {
#if YYDEBUG
                if (yydebug)
                    printf("%sdebug: state %d, error recovery shifting\
 to state %d\n", YYPREFIX, *yyssp, yytable[yyn]);
#endif
                if (yyssp >= yyss + yystacksize - 1)
                {
                    goto yyoverflow;
                }
                *++yyssp = yystate = yytable[yyn];
                *++yyvsp = yylval;
                goto yyloop;
            }
            else
            {
#if YYDEBUG
                if (yydebug)
                    printf("%sdebug: error recovery discarding state %d\n",
                            YYPREFIX, *yyssp);
#endif
                if (yyssp <= yyss) goto yyabort;
                --yyssp;
                --yyvsp;
            }
        }
    }
    else
    {
        if (yychar == 0) goto yyabort;
#if YYDEBUG
        if (yydebug)
        {
            yys = 0;
            if (yychar <= YYMAXTOKEN) yys = yyname[yychar];
            if (!yys) yys = "illegal-symbol";
            printf("%sdebug: state %d, error recovery discards token %d (%s)\n",
                    YYPREFIX, yystate, yychar, yys);
        }
#endif
        yychar = (-1);
        goto yyloop;
    }
yyreduce:
#if YYDEBUG
    if (yydebug)
        printf("%sdebug: state %d, reducing by rule %d (%s)\n",
                YYPREFIX, yystate, yyn, yyrule[yyn]);
#endif
    yym = yylen[yyn];
    yyval = yyvsp[1-yym];
    switch (yyn)
    {
case 5:
#line 206 "gram.y"
{ yyerrok; }
break;
case 6:
#line 208 "gram.y"
{ yyerrok; }
break;
case 7:
#line 210 "gram.y"
{ yyerrok; }
break;
case 8:
#line 218 "gram.y"
{ ignorenl++; }
break;
case 9:
#line 221 "gram.y"
{ ignorenl--; }
break;
case 10:
#line 224 "gram.y"
{ 
			ignorenl = 0; 
			LexNamespace = 0; 
			CurrentNamespace = TopNamespace; 
			funcDepth = 0; 
			notCommand = 0;
		    }
break;
case 11:
#line 241 "gram.y"
{ notCommand = 1; }
break;
case 18:
#line 256 "gram.y"
{
			ENTER ();
			ExprPtr	e;
			Value	t;
			NamespacePtr	s;
			FramePtr	f;
			CodePtr		c;
			
			e = BuildCall ("Command", "display",
				       1,NewExprTree (EXPR, yyvsp[-2].expr, 0));
			GetNamespace (&s, &f, &c);
			t = NewThread (f, CompileExpr (e, c));
			ThreadsRun (t, 0);
			EXIT ();
		    }
break;
case 19:
#line 272 "gram.y"
{
			ENTER ();
			ExprPtr	e;
			Value	t;
			NamespacePtr	s;
			FramePtr	f;
			CodePtr		c;

			e = BuildCall("Command", "display_base",
				      2, NewExprTree (EXPR, yyvsp[-4].expr, 0), yyvsp[-2].expr);
			GetNamespace (&s, &f, &c);
			t = NewThread (f, CompileExpr (e, c));
			ThreadsRun (t, 0);
			EXIT ();
		    }
break;
case 20:
#line 288 "gram.y"
{ 
			ENTER ();
			NamespacePtr    s;
			FramePtr	f;
			CodePtr		c;
			Value		t;
			
			GetNamespace (&s, &f, &c);
			t = NewThread (f, CompileStat (yyvsp[-2].expr, c));
			ThreadsRun (t, 0);
			EXIT ();
		    }
break;
case 21:
#line 301 "gram.y"
{
			ENTER();
			ExprPtr	e;
			Value	t;
			NamespacePtr	s;
			FramePtr	f;
			CodePtr		c;
			CommandPtr	cmd;

			cmd = CommandFind (CurrentCommands, yyvsp[-4].atom);
			if (!cmd)
			    ParseError ("Undefined command \"%s\"", AtomName (yyvsp[-4].atom));
			else
			{
			    e = NewExprTree (OP, 
					     NewExprConst (POLY_CONST, cmd->func),
					     yyvsp[-3].expr);
			    GetNamespace (&s, &f, &c);
			    t = NewThread (f, CompileExpr (e, c));
			    ThreadsRun (t, 0);
			}
			EXIT ();
		    }
break;
case 22:
#line 325 "gram.y"
{
			ENTER ();
			ExprPtr e;

			Value	t;
			NamespacePtr	s;
			FramePtr	f;
			CodePtr		c;
			CommandPtr	cmd;

			cmd = CommandFind (CurrentCommands, yyvsp[-4].atom);
			if (!cmd)
			    ParseError ("Undefined command \"%s\"", AtomName (yyvsp[-4].atom));
			else
			{
			    e = NewExprTree (OP, 
					     NewExprConst (POLY_CONST, cmd->func),
					     yyvsp[-3].expr);
			    GetNamespace (&s, &f, &c);
			    t = NewThread (f, CompileExpr (e, c));
			    ThreadsRun (t, 0);
			}
			EXIT ();
		    }
break;
case 26:
#line 354 "gram.y"
{ yyval.expr = 0; }
break;
case 27:
#line 357 "gram.y"
{ yyval.expr = NewExprTree (COMMA, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 28:
#line 359 "gram.y"
{ yyval.expr = NewExprTree (COMMA, yyvsp[0].expr, 0); }
break;
case 29:
#line 362 "gram.y"
{ yyval.expr = BuildRawname (yyvsp[-1].expr, yyvsp[0].atom); }
break;
case 33:
#line 369 "gram.y"
{ yyval.expr = NewExprTree (COLONCOLON, yyvsp[-2].expr, NewExprAtom (yyvsp[-1].atom, 0, False)); }
break;
case 34:
#line 371 "gram.y"
{ yyval.expr = 0; }
break;
case 35:
#line 374 "gram.y"
{ 
			yyval.expr = BuildFullname (yyvsp[-1].expr, yyvsp[0].atom);
			LexNamespace = 0;
		    }
break;
case 36:
#line 379 "gram.y"
{ 
			yyval.expr = BuildFullname (0, yyvsp[0].atom);
			LexNamespace = 0;
		    }
break;
case 37:
#line 385 "gram.y"
{ 
			yyval.expr = BuildFullname (yyvsp[-1].expr, yyvsp[0].atom);
			LexNamespace = 0;
		    }
break;
case 38:
#line 390 "gram.y"
{
			yyval.expr = BuildFullname (0, yyvsp[0].atom);
			LexNamespace = 0;
		    }
break;
case 39:
#line 396 "gram.y"
{ 
			ExprPtr	    e;
			SymbolPtr   symbol;

			e = BuildFullname (yyvsp[-2].expr, yyvsp[-1].atom);
			if (e->base.tag == COLONCOLON)
			    symbol = e->tree.right->atom.symbol;
			else
			    symbol = e->atom.symbol;
			if (!symbol)
			{
			    ParseError ("non-existant namespace \"%A\"", yyvsp[-1].atom);
			    YYERROR;
			}
			else if (symbol->symbol.class != class_namespace)
			{
			    ParseError ("%A is not a namespace", yyvsp[-1].atom);
			    YYERROR;
			}
			LexNamespace = symbol->namespace.namespace;
			yyval.expr = e;
		    }
break;
case 40:
#line 419 "gram.y"
{ 
			ExprPtr	    e;
			SymbolPtr   symbol;

			e = BuildFullname (0, yyvsp[-1].atom);
			if (e->base.tag == COLONCOLON)
			    symbol = e->tree.right->atom.symbol;
			else
			    symbol = e->atom.symbol;
			if (!symbol)
			{
			    ParseError ("non-existant namespace \"%A\"", yyvsp[-1].atom);
			    YYERROR;
			}
			else if (symbol->symbol.class != class_namespace)
			{
			    ParseError ("%A is not a namespace", yyvsp[-1].atom);
			    YYERROR;
			}
			LexNamespace = symbol->namespace.namespace;
			yyval.expr = e;
		    }
break;
case 43:
#line 449 "gram.y"
{ CurrentNamespace = NewNamespace (CurrentNamespace); }
break;
case 44:
#line 452 "gram.y"
{ CurrentNamespace = CurrentNamespace->previous; }
break;
case 45:
#line 455 "gram.y"
{ yyval.expr = yyvsp[-1].expr; }
break;
case 48:
#line 462 "gram.y"
{ yyval.expr = NewExprTree(OC, yyvsp[-1].expr, yyvsp[0].expr); }
break;
case 49:
#line 464 "gram.y"
{ yyval.expr = NewExprTree(OC, 0, 0); }
break;
case 50:
#line 467 "gram.y"
{ yyval.expr = NewExprTree(IF, yyvsp[-4].expr, yyvsp[-2].expr); }
break;
case 51:
#line 469 "gram.y"
{ yyval.expr = NewExprTree(ELSE, yyvsp[-6].expr, NewExprTree(ELSE, yyvsp[-4].expr, yyvsp[-2].expr)); }
break;
case 52:
#line 471 "gram.y"
{ yyval.expr = NewExprTree(WHILE, yyvsp[-4].expr, yyvsp[-2].expr); }
break;
case 53:
#line 473 "gram.y"
{ yyval.expr = NewExprTree(DO, yyvsp[-6].expr, yyvsp[-3].expr); }
break;
case 54:
#line 475 "gram.y"
{ Expr *expr = ParseCanonFor(yyvsp[-4].expr);
		      if (!expr)
                          YYERROR;
		      yyval.expr = NewExprTree(FOR, expr, yyvsp[-2].expr); }
break;
case 55:
#line 480 "gram.y"
{ yyval.expr = NewExprTree (SWITCH, yyvsp[-4].expr, yyvsp[-2].expr); }
break;
case 56:
#line 482 "gram.y"
{ yyval.expr = NewExprTree (UNION, yyvsp[-4].expr, yyvsp[-2].expr); }
break;
case 57:
#line 484 "gram.y"
{ yyval.expr = NewExprTree(BREAK, (Expr *) 0, (Expr *) 0); }
break;
case 58:
#line 486 "gram.y"
{ yyval.expr = NewExprTree(CONTINUE, (Expr *) 0, (Expr *) 0); }
break;
case 59:
#line 488 "gram.y"
{ yyval.expr = NewExprTree (RETURNTOK, (Expr *) 0, yyvsp[-1].expr); }
break;
case 60:
#line 490 "gram.y"
{ yyval.expr = NewExprTree(EXPR, yyvsp[-1].expr, (Expr *) 0); }
break;
case 62:
#line 493 "gram.y"
{ yyval.expr = NewExprTree(SEMI, (Expr *) 0, (Expr *) 0); }
break;
case 63:
#line 495 "gram.y"
{ 
			DeclList    *decl = yyvsp[-3].funcDecl.decl;
			SymbolPtr   symbol = decl->symbol;
			Class	    class = yyvsp[-3].funcDecl.type.class;
			Publish	    publish = yyvsp[-3].funcDecl.type.publish;
			TypePtr	    type = yyvsp[-3].funcDecl.type.type;
			TypePtr	    ret = type->func.ret;
			ArgType	    *argType = type->func.args;

			if (symbol)
			{
			    if (yyvsp[-1].expr)
			    {
				symbol->symbol.forward = False;
				ParseCanonType (ret, False);
				decl->init = NewExprCode (NewFuncCode (ret,
								       argType,
								       yyvsp[-1].expr,
								       yyvsp[-2].value),
							  NewExprAtom (symbol->symbol.name, symbol, False));
			    }
			    else
				symbol->symbol.forward = True;
			}
			yyval.expr = NewExprDecl (FUNC, decl, class, type, publish);
		    }
break;
case 64:
#line 522 "gram.y"
{ 
			DeclListPtr decl;

			decl = NewDeclList (yyvsp[-6].atom, 0, 0);
			decl->symbol = ParseNewSymbol (yyvsp[-9].publish, 
						       class_exception, 
						       typePoly, yyvsp[-6].atom);
			decl->symbol->exception.doc = yyvsp[-2].value;
			yyval.expr = NewExprDecl (EXCEPTION,
					  decl,
					  class_exception,
					  NewTypeFunc (typePoly, yyvsp[-4].argType),
					  yyvsp[-9].publish);
		    }
break;
case 65:
#line 537 "gram.y"
{ yyval.expr = NewExprTree (RAISE, yyvsp[-4].expr, yyvsp[-2].expr); }
break;
case 66:
#line 539 "gram.y"
{ 
			DeclListPtr decl;

			for (decl = yyvsp[-2].declList; decl; decl = decl->next)
			    decl->symbol = ParseNewSymbol (yyvsp[-5].publish, class_typedef,
							   0, decl->name);
			yyval.expr = NewExprTree (TYPEDEF, NewExprDecl (TYPEDEF, yyvsp[-2].declList, class_typedef, 0, yyvsp[-5].publish), 0);
		    }
break;
case 67:
#line 548 "gram.y"
{ 
			DeclListPtr decl;

			for (decl = yyvsp[-2].declList; decl; decl = decl->next)
			    decl->symbol = ParseNewSymbol (yyvsp[-6].publish, class_typedef,
							   yyvsp[-3].type, decl->name);
			yyval.expr = NewExprTree (TYPEDEF, NewExprDecl (TYPEDEF, yyvsp[-2].declList, class_typedef, yyvsp[-3].type, yyvsp[-6].publish), 0);
		    }
break;
case 68:
#line 557 "gram.y"
{
			SymbolPtr   symbol;
			Publish	    publish = yyvsp[-3].publish;
			
			/*
			 * this is a hack - save the current namespace
			 * on the parser stack to be restored after
			 * the block is compiled.
			 */
			yyvsp[-2].namespace = CurrentNamespace;
			if (publish == publish_extend)
			{
			    symbol = NamespaceFindName (CurrentNamespace, yyvsp[0].atom, True);
			    if (!symbol)
			    {
				ParseError ("non-existant namespace %A", yyvsp[0].atom);
				YYERROR;
			    }
			    else if (symbol->symbol.class != class_namespace)
			    {
				ParseError ("%A is not a namespace", yyvsp[0].atom);
				YYERROR;
			    }
			    else
				CurrentNamespace = symbol->namespace.namespace;
			}
			else
			{
			    symbol = ParseNewSymbol (yyvsp[-3].publish, class_namespace,
						     0, yyvsp[0].atom);
			    CurrentNamespace = NewNamespace (CurrentNamespace);
			    symbol->namespace.namespace = CurrentNamespace;
			}
			/*
			 * Make all of the symbols visible while compiling within
			 * the namespace
			 */
			if (CurrentNamespace != yyvsp[-2].namespace)
			    CurrentNamespace->publish = publish_public;
		    }
break;
case 69:
#line 598 "gram.y"
{
			/*
			 * close the namespace to non-public lookups
			 */
			if (CurrentNamespace != yyvsp[-7].namespace)
			    CurrentNamespace->publish = publish_private;
			/*
			 * Restore to the namespace saved on
			 * the parser stack
			 */
			CurrentNamespace = yyvsp[-7].namespace;
			yyval.expr = NewExprTree (NAMESPACE, NewExprAtom (yyvsp[-5].atom, 0, False), yyvsp[-2].expr);
		    }
break;
case 70:
#line 612 "gram.y"
{
			SymbolPtr	symbol;
			ExprPtr		e;

			e = yyvsp[-2].expr;
			if (yyvsp[-2].expr->base.tag == COLONCOLON)
			    e = e->tree.right;
			symbol = e->atom.symbol;
			if (!symbol)
			{
			    ParseError ("non-existant namespace %A", e->atom.atom);
			    YYERROR;
			}
			else if (symbol->symbol.class != class_namespace)
			{
			    ParseError ("%A is not a namespace", e->atom.atom);
			    YYERROR;
			}
			NamespaceImport (CurrentNamespace, 
					 symbol->namespace.namespace, yyvsp[-5].publish);
			yyval.expr = NewExprTree (IMPORT, yyvsp[-2].expr, NewExprDecl (IMPORT,
								   0, 
								   class_undef,
								   0,
								   yyvsp[-5].publish));
		    }
break;
case 71:
#line 639 "gram.y"
{ yyval.expr = NewExprTree (CATCH, yyvsp[-1].expr, yyvsp[-2].expr); }
break;
case 72:
#line 641 "gram.y"
{ yyval.expr = NewExprTree (TWIXT, 
					NewExprTree (TWIXT, yyvsp[-6].expr, yyvsp[-4].expr),
					NewExprTree (TWIXT, yyvsp[-2].expr, 0));
		    }
break;
case 73:
#line 647 "gram.y"
{ yyval.expr = NewExprTree(SEMI, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 74:
#line 649 "gram.y"
{ yyval.expr = NewExprTree(SEMI, yyvsp[0].expr, 0); }
break;
case 75:
#line 652 "gram.y"
{ yyval.expr = NewExprTree (CATCH, yyvsp[-1].expr, yyvsp[0].expr); }
break;
case 76:
#line 654 "gram.y"
{ yyval.expr = 0; }
break;
case 77:
#line 657 "gram.y"
{ yyval.expr = NewExprCode (NewFuncCode (typePrim[rep_void],
						     yyvsp[-3].argType, yyvsp[-1].expr, yyvsp[-2].value),
					yyvsp[-5].expr); 
		    }
break;
case 79:
#line 664 "gram.y"
{ yyval.expr = 0; }
break;
case 80:
#line 666 "gram.y"
{ ++funcDepth; }
break;
case 81:
#line 666 "gram.y"
{ --funcDepth; yyval.expr = yyvsp[0].expr; }
break;
case 82:
#line 669 "gram.y"
{ yyval.expr = yyvsp[0].expr; }
break;
case 83:
#line 671 "gram.y"
{ 
			yyval.expr = NewExprTree (OC,
					  NewExprTree (RETURNTOK, 0, yyvsp[-1].expr),
					  NewExprTree (OC, 0, 0));
		    }
break;
case 84:
#line 678 "gram.y"
{ seeComment = 1; }
break;
case 85:
#line 681 "gram.y"
{ seeComment = 0; yyval.value = yyvsp[0].value; }
break;
case 87:
#line 685 "gram.y"
{ yyval.value = Void; }
break;
case 88:
#line 688 "gram.y"
{ yyval.expr = yyvsp[-1].expr; }
break;
case 89:
#line 691 "gram.y"
{ yyval.expr = NewExprTree (CASE, yyvsp[-1].expr, yyvsp[0].expr); }
break;
case 90:
#line 693 "gram.y"
{ yyval.expr = 0; }
break;
case 91:
#line 696 "gram.y"
{ yyval.expr = NewExprTree (CASE, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 92:
#line 698 "gram.y"
{ yyval.expr = NewExprTree (CASE, 0, yyvsp[0].expr); }
break;
case 93:
#line 701 "gram.y"
{ yyval.expr = yyvsp[-1].expr; }
break;
case 94:
#line 704 "gram.y"
{ yyval.expr = NewExprTree (CASE, yyvsp[-1].expr, yyvsp[0].expr); }
break;
case 95:
#line 706 "gram.y"
{ yyval.expr = 0; }
break;
case 96:
#line 709 "gram.y"
{ 
			SymbolPtr   sym = yyvsp[-3].symbol;
			Atom	    sym_atom = sym ? sym->symbol.name : 0;
			ExprPtr	    name = 0;
			
			if (sym)
			    name = NewExprAtom (sym_atom, sym, False);
			    
			yyval.expr = NewExprTree (CASE, 
					  NewExprTree (CASE,
						       NewExprAtom (yyvsp[-4].atom, 0, False),
						       name),
					  yyvsp[-1].expr);
		    }
break;
case 97:
#line 724 "gram.y"
{ yyval.expr = NewExprTree (CASE, 0, yyvsp[0].expr); }
break;
case 98:
#line 727 "gram.y"
{
			yyval.symbol = ParseNewSymbol (publish_private,
					     class_undef,
					     typePoly,
					     yyvsp[0].atom);
		    }
break;
case 99:
#line 734 "gram.y"
{ yyval.symbol = 0; }
break;
case 100:
#line 740 "gram.y"
{ yyval.atomList = NewAtomList (yyvsp[0].atomList, yyvsp[-2].atom); }
break;
case 101:
#line 742 "gram.y"
{ yyval.atomList = NewAtomList (0, yyvsp[0].atom); }
break;
case 102:
#line 745 "gram.y"
{ yyval.declList = NewDeclList (yyvsp[-2].atom, 0, yyvsp[0].declList); }
break;
case 103:
#line 747 "gram.y"
{ yyval.declList = NewDeclList (yyvsp[0].atom, 0, 0); }
break;
case 106:
#line 760 "gram.y"
{ 
			if (yyvsp[-3].symbol)
			{
			    yyval.declList = NewDeclList (yyvsp[-3].symbol->symbol.name, yyvsp[-2].expr, yyvsp[0].declList); 
			    yyval.declList->symbol = yyvsp[-3].symbol;
			}
			else
			    yyval.declList = yyvsp[0].declList;
		    }
break;
case 107:
#line 770 "gram.y"
{ 
			if (yyvsp[-1].symbol)
			{
			    yyval.declList = NewDeclList (yyvsp[-1].symbol->symbol.name, yyvsp[0].expr, 0);
			    yyval.declList->symbol = yyvsp[-1].symbol;
			}
			else
			    yyvsp[-1].symbol = 0;
		    }
break;
case 108:
#line 781 "gram.y"
{	
			yyval.symbol = ParseNewSymbol (yyvsp[-1].fulltype.publish,
					     yyvsp[-1].fulltype.class,
					     yyvsp[-1].fulltype.type,
					     yyvsp[0].atom);
		    }
break;
case 109:
#line 793 "gram.y"
{ yyval.fulltype = yyvsp[-3].fulltype; }
break;
case 110:
#line 799 "gram.y"
{
			NamespacePtr	save = CurrentNamespace;
			SymbolPtr	symbol;
			Type		*type = NewTypeFunc (yyvsp[-3].funcDecl.type.type, yyvsp[-1].argType);
			/*
			 * namespace_start pushed a new namespace, make sure
			 * this symbol is placed in the original namespace
			 */
			CurrentNamespace = save->previous;
			symbol = NamespaceFindName (CurrentNamespace, yyvsp[-3].funcDecl.decl->name, True);
			if (symbol && symbol->symbol.forward)
			{
			    if (!TypeIsSupertype (symbol->symbol.type, type))
			    {
				ParseError ("%A redefinition with different type",
					    yyvsp[-3].funcDecl.decl->name);
				symbol = 0;
			    }
			}
			else
			{
			    symbol = ParseNewSymbol (yyvsp[-3].funcDecl.type.publish,
						     yyvsp[-3].funcDecl.type.class, 
						     type,
						     yyvsp[-3].funcDecl.decl->name);
			}
			CurrentNamespace = save;
			yyval.funcDecl = yyvsp[-3].funcDecl;
			yyval.funcDecl.decl->symbol = symbol;
			yyval.funcDecl.type.type = type;
		    }
break;
case 111:
#line 832 "gram.y"
{ yyval.funcDecl.type = yyvsp[-2].fulltype; yyval.funcDecl.decl = NewDeclList (yyvsp[-1].atom, 0, 0); }
break;
case 112:
#line 834 "gram.y"
{ yyval.funcDecl.type = yyvsp[-3].fulltype; yyval.funcDecl.decl = NewDeclList (yyvsp[-1].atom, 0, 0); }
break;
case 113:
#line 836 "gram.y"
{ 
			yyval.funcDecl.type.publish = publish_private;
			yyval.funcDecl.type.class = class_undef;
			yyval.funcDecl.type.type = typePoly;
			yyval.funcDecl.decl = NewDeclList (yyvsp[-1].atom, 0, 0); 
		    }
break;
case 114:
#line 844 "gram.y"
{ yyval.expr = yyvsp[0].expr; }
break;
case 115:
#line 846 "gram.y"
{ yyval.expr = yyvsp[0].expr; }
break;
case 116:
#line 848 "gram.y"
{ yyval.expr = 0; }
break;
case 117:
#line 854 "gram.y"
{ yyval.fulltype.publish = yyvsp[-3].publish; yyval.fulltype.class = yyvsp[-2].class; yyval.fulltype.type = yyvsp[-1].type; }
break;
case 118:
#line 856 "gram.y"
{ yyval.fulltype.publish = publish_private; yyval.fulltype.class = yyvsp[-2].class; yyval.fulltype.type = yyvsp[-1].type; }
break;
case 119:
#line 858 "gram.y"
{ yyval.fulltype.publish = yyvsp[-2].publish; yyval.fulltype.class = class_undef; yyval.fulltype.type = yyvsp[-1].type; }
break;
case 120:
#line 860 "gram.y"
{ yyval.fulltype.publish = publish_private; yyval.fulltype.class = class_undef; yyval.fulltype.type = yyvsp[-1].type; }
break;
case 121:
#line 862 "gram.y"
{ yyval.fulltype.publish = yyvsp[-2].publish; yyval.fulltype.class = yyvsp[-1].class; yyval.fulltype.type = typePoly; }
break;
case 122:
#line 864 "gram.y"
{ yyval.fulltype.publish = publish_private; yyval.fulltype.class = yyvsp[-1].class; yyval.fulltype.type = typePoly; }
break;
case 123:
#line 866 "gram.y"
{ yyval.fulltype.publish = yyvsp[-1].publish; yyval.fulltype.class = class_undef; yyval.fulltype.type = typePoly; }
break;
case 124:
#line 872 "gram.y"
{ yyval.type = NewTypeFunc (yyvsp[0].type, yyvsp[-1].argType); }
break;
case 125:
#line 874 "gram.y"
{ yyval.type = NewTypeArray (yyvsp[0].type, yyvsp[-2].expr, False); }
break;
case 126:
#line 876 "gram.y"
{ yyval.type = NewTypeArray (yyvsp[0].type, yyvsp[-2].expr, True); }
break;
case 127:
#line 878 "gram.y"
{ yyval.type = NewTypeArray (yyvsp[0].type, yyvsp[-2].expr, False); }
break;
case 128:
#line 880 "gram.y"
{ yyval.type = NewTypeHash (yyvsp[0].type, yyvsp[-2].type); }
break;
case 129:
#line 882 "gram.y"
{ yyval.type = 0; }
break;
case 130:
#line 885 "gram.y"
{
			Type	*top = yyvsp[0].type;
			Type	*t, **bot = &top;

			/*
			 * Walk down the type chain to hang the
			 * base type off of the end.  This
			 * makes int[]() be an array of functions
			 * rather than a function returning an array
			 */
			while ((t = *bot))
			    switch (t->base.tag) {
			    case type_array:
				bot = &t->array.type;
				break;
			    case type_hash:
				bot = &t->hash.type;
				break;
			    case type_func:
				bot = &t->func.ret;
				break;
			    default:
				assert(0);
			    }
			*bot = yyvsp[-1].type;
			yyval.type = top;
		    }
break;
case 131:
#line 913 "gram.y"
{ yyval.type = NewTypeRef (yyvsp[0].type, True); }
break;
case 132:
#line 915 "gram.y"
{ yyval.type = NewTypeRef (NewTypeRef (yyvsp[0].type, True), True); }
break;
case 133:
#line 917 "gram.y"
{ yyval.type = NewTypeRef (NewTypeRef (yyvsp[0].type, False), False); }
break;
case 134:
#line 919 "gram.y"
{ yyval.type = NewTypeRef (yyvsp[0].type, False); }
break;
case 136:
#line 923 "gram.y"
{
			AtomListPtr	al;
			StructType	*st;
			MemListPtr	ml;
			int		nelements;

			nelements = 0;
			for (ml = yyvsp[-1].memList; ml; ml = ml->next)
			{
			    for (al = ml->atoms; al; al = al->next)
				nelements++;
			}
			st = NewStructType (nelements);
			nelements = 0;
			for (ml = yyvsp[-1].memList; ml; ml = ml->next)
			{
			    for (al = ml->atoms; al; al = al->next)
			    {
				AddBoxType (&st->types, ml->type);
				StructTypeAtoms(st)[nelements] = al->atom;
				nelements++;
			    }
			}
			yyval.type = NewTypeStruct (st);
		    }
break;
case 137:
#line 949 "gram.y"
{
			AtomListPtr	al;
			StructType	*st;
			MemListPtr	ml;
			int		nelements;

			nelements = 0;
			for (ml = yyvsp[-1].memList; ml; ml = ml->next)
			{
			    for (al = ml->atoms; al; al = al->next)
				nelements++;
			}
			st = NewStructType (nelements);
			nelements = 0;
			for (ml = yyvsp[-1].memList; ml; ml = ml->next)
			{
			    for (al = ml->atoms; al; al = al->next)
			    {
				AddBoxType (&st->types, ml->type);
				StructTypeAtoms(st)[nelements] = al->atom;
				nelements++;
			    }
			}
			yyval.type = NewTypeUnion (st, False);
		    }
break;
case 138:
#line 975 "gram.y"
{
			AtomListPtr	al;
			StructType	*st;
			int		nelements;

			nelements = 0;
			for (al = yyvsp[-1].atomList; al; al = al->next)
			    nelements++;
			
			st = NewStructType (nelements);
			nelements = 0;
			for (al = yyvsp[-1].atomList; al; al = al->next)
			{
			    AddBoxType (&st->types, typePrim[rep_void]);
			    StructTypeAtoms(st)[nelements] = al->atom;
			    nelements++;
			}
			yyval.type = NewTypeUnion (st, True);
		    }
break;
case 139:
#line 996 "gram.y"
{ yyval.type = yyvsp[-1].type; }
break;
case 140:
#line 998 "gram.y"
{ yyval.type = NewTypeName (yyvsp[0].expr, 0); }
break;
case 142:
#line 1002 "gram.y"
{ yyval.type = typePoly; }
break;
case 156:
#line 1019 "gram.y"
{ yyval.expr = 0; }
break;
case 157:
#line 1022 "gram.y"
{ yyval.expr = NewExprTree (COMMA, 0, yyvsp[-2].expr); }
break;
case 158:
#line 1024 "gram.y"
{ yyval.expr = NewExprTree (COMMA, 0, 0); }
break;
case 159:
#line 1027 "gram.y"
{ yyval.expr = NewExprTree (COMMA, 0, yyvsp[-2].expr); }
break;
case 160:
#line 1029 "gram.y"
{ yyval.expr = NewExprTree (COMMA, 0, 0); }
break;
case 161:
#line 1032 "gram.y"
{ yyval.expr = NewExprTree (COMMA, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 162:
#line 1034 "gram.y"
{ yyval.expr = NewExprTree (COMMA, yyvsp[0].expr, 0); }
break;
case 163:
#line 1040 "gram.y"
{ yyval.memList = NewMemList (yyvsp[-2].atomList, yyvsp[-3].type, yyvsp[0].memList); }
break;
case 164:
#line 1042 "gram.y"
{ yyval.memList = 0; }
break;
case 165:
#line 1045 "gram.y"
{ yyval.memList = NewMemList (yyvsp[-2].atomList, yyvsp[-3].type, yyvsp[0].memList); }
break;
case 166:
#line 1047 "gram.y"
{ yyval.memList = 0; }
break;
case 172:
#line 1060 "gram.y"
{ yyval.publish = publish_private; }
break;
case 177:
#line 1074 "gram.y"
{ yyval.argType = yyvsp[-1].argType; }
break;
case 178:
#line 1076 "gram.y"
{ yyval.argType = 0; }
break;
case 179:
#line 1079 "gram.y"
{ yyval.argType = NewArgType (yyvsp[-2].argDecl.type, False, yyvsp[-2].argDecl.name, 0, yyvsp[0].argType); }
break;
case 180:
#line 1081 "gram.y"
{ yyval.argType = NewArgType (yyvsp[-1].argDecl.type, yyvsp[0].bool, yyvsp[-1].argDecl.name, 0, 0); }
break;
case 181:
#line 1084 "gram.y"
{ 
			ParseCanonType (yyvsp[-1].type, False);
			yyval.argDecl.type = yyvsp[-1].type; 
			yyval.argDecl.name = yyvsp[0].atom; 
		    }
break;
case 182:
#line 1090 "gram.y"
{ 
			ParseCanonType (yyvsp[0].type, False);
			yyval.argDecl.type = yyvsp[0].type;
			yyval.argDecl.name = 0; 
		    }
break;
case 183:
#line 1101 "gram.y"
{ yyval.argType = yyvsp[-1].argType; }
break;
case 184:
#line 1104 "gram.y"
{
			ArgType	*args;
			Type	*type;

			for (args = yyvsp[0].argType; args; args = args->next)
			{
			    type = args->type;
			    if (ParseCanonType (type, False) != CanonTypeDefined)
				break;
			    if (args->varargs)
			    {
				type = NewTypeArray (type,
						     NewExprTree (COMMA,
								  NewExprConst (TEN_NUM, 
										NewInt (0)),
								  0),
						     False);
			    }
			    args->symbol = ParseNewSymbol (publish_private,
							   class_arg, 
							   type, args->name);
			}
			yyval.argType = yyvsp[0].argType;
		    }
break;
case 185:
#line 1129 "gram.y"
{ yyval.argType = 0; }
break;
case 186:
#line 1132 "gram.y"
{ yyval.argType = NewArgType (yyvsp[-2].argDecl.type, False, yyvsp[-2].argDecl.name, 0, yyvsp[0].argType); }
break;
case 187:
#line 1134 "gram.y"
{ yyval.argType = NewArgType (yyvsp[-1].argDecl.type, yyvsp[0].bool, yyvsp[-1].argDecl.name, 0, 0); }
break;
case 188:
#line 1137 "gram.y"
{ yyval.argDecl.type = yyvsp[-1].type; yyval.argDecl.name = yyvsp[0].atom; }
break;
case 189:
#line 1139 "gram.y"
{ yyval.argDecl.type = yyvsp[0].type; yyval.argDecl.name = 0; }
break;
case 190:
#line 1142 "gram.y"
{ yyval.bool = True; }
break;
case 191:
#line 1144 "gram.y"
{ yyval.bool = False; }
break;
case 193:
#line 1152 "gram.y"
{ yyval.expr = 0; }
break;
case 195:
#line 1156 "gram.y"
{ 
			DeclList    *decl;

			for (decl = yyvsp[0].declList; decl; decl = decl->next)
			{
			    if (decl->init)
			    {
				if (!decl->init->base.type)
				    decl->init->base.type = yyvsp[-1].fulltype.type;
			    }
			}

			yyval.expr = NewExprDecl (VAR, yyvsp[0].declList, yyvsp[-1].fulltype.class, yyvsp[-1].fulltype.type, yyvsp[-1].fulltype.publish); 
		    }
break;
case 197:
#line 1173 "gram.y"
{ yyval.expr = NewExprTree(COMMA, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 199:
#line 1180 "gram.y"
{ yyval.expr = 0; }
break;
case 200:
#line 1183 "gram.y"
{ yyval.expr = NewExprTree (COMMA, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 201:
#line 1185 "gram.y"
{ yyval.expr = NewExprTree (COMMA, yyvsp[0].expr, 0); }
break;
case 203:
#line 1189 "gram.y"
{ yyval.expr = 0; }
break;
case 204:
#line 1192 "gram.y"
{ yyval.expr = NewExprTree (COMMA, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 205:
#line 1194 "gram.y"
{	
			ExprPtr	arg = yyvsp[0].bool ? NewExprTree (DOTDOTDOT, yyvsp[-1].expr, 0) : yyvsp[-1].expr;
			yyval.expr = NewExprTree (COMMA, arg, 0); 
		    }
break;
case 206:
#line 1200 "gram.y"
{ 
			yyval.expr = NewExprTree (OC,
					  NewExprTree (RETURNTOK, 0, yyvsp[0].expr),
					  NewExprTree (OC, 0, 0));
		    }
break;
case 207:
#line 1205 "gram.y"
{ ++funcDepth; }
break;
case 208:
#line 1205 "gram.y"
{ --funcDepth; yyval.expr = yyvsp[0].expr; }
break;
case 209:
#line 1211 "gram.y"
{ 
			if (yyvsp[-1].ints == ASSIGNPOW)
			    yyval.expr = NewExprTree (ASSIGNPOW, 
					      BuildName ("Math", "assign_pow"),
					      NewExprTree (ASSIGNPOW, yyvsp[-2].expr, yyvsp[0].expr));
			else
			{
			    ExprPtr left = yyvsp[-2].expr;
			    /*
			     * Automatically declare names used in
			     * simple assignements at the top level
			     */
			    if (yyvsp[-1].ints == ASSIGN && 
				funcDepth == 0 &&
				left->base.tag == NAME && 
				!(left->atom.symbol))
			    {
				yyvsp[-2].expr->atom.symbol = ParseNewSymbol (publish_private,
								  class_undef, 
								  typePoly, 
								  yyvsp[-2].expr->atom.atom);
			    }
			    yyval.expr = NewExprTree(yyvsp[-1].ints, yyvsp[-2].expr, yyvsp[0].expr); 
			}
		    }
break;
case 210:
#line 1237 "gram.y"
{
			ParseCanonType (yyvsp[-6].type, False);
			yyval.expr = NewExprCode (NewFuncCode (yyvsp[-6].type, yyvsp[-3].argType, yyvsp[-1].expr, yyvsp[-2].value), 0); 
		    }
break;
case 211:
#line 1242 "gram.y"
{   Value	t;
			t = do_Thread_id_to_thread (yyvsp[0].value);
			if (t == Void)
			{
			    ParseError ("No thread %v", yyvsp[0].value);
			    YYERROR;
			}
			else
			    yyval.expr = NewExprConst(THREAD_CONST, t); 
		    }
break;
case 212:
#line 1253 "gram.y"
{ yyval.expr = NewExprTree(STAR, yyvsp[0].expr, (Expr *) 0); }
break;
case 213:
#line 1255 "gram.y"
{ yyval.expr = NewExprTree(STAR, NewExprTree (STAR, yyvsp[0].expr, 0), 0); }
break;
case 214:
#line 1257 "gram.y"
{ yyval.expr = NewExprTree(AMPER, yyvsp[0].expr, (Expr *) 0); }
break;
case 215:
#line 1259 "gram.y"
{ yyval.expr = NewExprTree(AMPER, 
				       NewExprTree (AMPER, yyvsp[0].expr, (Expr *) 0), 
				       (Expr *) 0); }
break;
case 216:
#line 1263 "gram.y"
{ yyval.expr = NewExprTree(UMINUS, yyvsp[0].expr, (Expr *) 0); }
break;
case 217:
#line 1265 "gram.y"
{ yyval.expr = NewExprTree(LNOT, yyvsp[0].expr, (Expr *) 0); }
break;
case 218:
#line 1267 "gram.y"
{ yyval.expr = NewExprTree(BANG, yyvsp[0].expr, (Expr *) 0); }
break;
case 219:
#line 1269 "gram.y"
{ yyval.expr = NewExprTree(FACT, yyvsp[-1].expr, (Expr *) 0); }
break;
case 220:
#line 1271 "gram.y"
{ yyval.expr = NewExprTree(INC, yyvsp[0].expr, (Expr *) 0); }
break;
case 221:
#line 1273 "gram.y"
{ yyval.expr = NewExprTree(INC, (Expr *) 0, yyvsp[-1].expr); }
break;
case 222:
#line 1275 "gram.y"
{ yyval.expr = NewExprTree(DEC, yyvsp[0].expr, (Expr *) 0); }
break;
case 223:
#line 1277 "gram.y"
{ yyval.expr = NewExprTree(DEC, (Expr *) 0, yyvsp[-1].expr); }
break;
case 224:
#line 1279 "gram.y"
{ yyval.expr = NewExprTree (FORK, (Expr *) 0, yyvsp[0].expr); }
break;
case 225:
#line 1281 "gram.y"
{ yyval.expr = NewExprTree(PLUS, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 226:
#line 1283 "gram.y"
{ yyval.expr = NewExprTree(MINUS, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 227:
#line 1285 "gram.y"
{ yyval.expr = NewExprTree(TIMES, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 228:
#line 1287 "gram.y"
{ yyval.expr = NewExprTree(DIVIDE, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 229:
#line 1289 "gram.y"
{ yyval.expr = NewExprTree(DIV, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 230:
#line 1291 "gram.y"
{ yyval.expr = NewExprTree(MOD, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 231:
#line 1293 "gram.y"
{ 
			yyval.expr = NewExprTree(POW, 
					 BuildName ("Math", "pow"), 
					 NewExprTree (POW, yyvsp[-2].expr, yyvsp[0].expr)); 
		    }
break;
case 232:
#line 1299 "gram.y"
{ yyval.expr = NewExprTree(SHIFTL, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 233:
#line 1301 "gram.y"
{ yyval.expr = NewExprTree(SHIFTR, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 234:
#line 1303 "gram.y"
{ yyval.expr = NewExprTree(QUEST, yyvsp[-4].expr, NewExprTree(COLON, yyvsp[-2].expr, yyvsp[0].expr)); }
break;
case 235:
#line 1305 "gram.y"
{ yyval.expr = NewExprTree(LXOR, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 236:
#line 1307 "gram.y"
{ yyval.expr = NewExprTree(LAND, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 237:
#line 1309 "gram.y"
{ yyval.expr = NewExprTree(LOR, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 238:
#line 1311 "gram.y"
{ yyval.expr = NewExprTree(AND, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 239:
#line 1313 "gram.y"
{ yyval.expr = NewExprTree(OR, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 240:
#line 1315 "gram.y"
{ yyval.expr = NewExprTree(EQ, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 241:
#line 1317 "gram.y"
{ yyval.expr = NewExprTree(NE, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 242:
#line 1319 "gram.y"
{ yyval.expr = NewExprTree(LT, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 243:
#line 1321 "gram.y"
{ yyval.expr = NewExprTree(GT, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 244:
#line 1323 "gram.y"
{ yyval.expr = NewExprTree(LE, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 245:
#line 1325 "gram.y"
{ yyval.expr = NewExprTree(GE, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 247:
#line 1328 "gram.y"
{ yyval.expr = NewExprConst(TEN_NUM, yyvsp[0].value); }
break;
case 248:
#line 1330 "gram.y"
{ yyval.expr = NewExprConst(OCTAL_NUM, yyvsp[0].value); }
break;
case 249:
#line 1332 "gram.y"
{ yyval.expr = NewExprConst(OCTAL0_NUM, yyvsp[0].value); }
break;
case 250:
#line 1334 "gram.y"
{ yyval.expr = NewExprConst(BINARY_NUM, yyvsp[0].value); }
break;
case 251:
#line 1336 "gram.y"
{ yyval.expr = NewExprConst(HEX_NUM, yyvsp[0].value); }
break;
case 252:
#line 1338 "gram.y"
{ yyval.expr = NewExprConst(TEN_FLOAT, yyvsp[0].value); }
break;
case 253:
#line 1340 "gram.y"
{ yyval.expr = NewExprConst(OCTAL_FLOAT, yyvsp[0].value); }
break;
case 254:
#line 1342 "gram.y"
{ yyval.expr = NewExprConst(OCTAL0_FLOAT, yyvsp[0].value); }
break;
case 255:
#line 1344 "gram.y"
{ yyval.expr = NewExprConst(BINARY_FLOAT, yyvsp[0].value); }
break;
case 256:
#line 1346 "gram.y"
{ yyval.expr = NewExprConst(HEX_FLOAT, yyvsp[0].value); }
break;
case 257:
#line 1348 "gram.y"
{ yyval.expr = NewExprConst(CHAR_CONST, yyvsp[0].value); }
break;
case 258:
#line 1350 "gram.y"
{ yyval.expr = NewExprConst(STRING_CONST, yyvsp[0].value); }
break;
case 259:
#line 1352 "gram.y"
{ yyval.expr = NewExprConst(VOIDVAL, yyvsp[0].value); }
break;
case 260:
#line 1354 "gram.y"
{ yyval.expr = NewExprConst(BOOLVAL, yyvsp[0].value); }
break;
case 261:
#line 1356 "gram.y"
{ 
			yyval.expr = BuildFullname (0, AtomId ("[]"));
		    }
break;
case 262:
#line 1360 "gram.y"
{ 
			ParseCanonType (yyvsp[-4].type, False);
			if (yyvsp[-1].expr)
			    yyvsp[-1].expr->base.type = yyvsp[-4].type;
			yyval.expr = NewExprTree (NEW, yyvsp[-1].expr, 0);
			yyval.expr->base.type = yyvsp[-4].type;
		    }
break;
case 263:
#line 1368 "gram.y"
{ 
			yyvsp[-1].expr->base.type = NewTypeArray (typePoly, yyvsp[-5].expr, True); 
			ParseCanonType (yyvsp[-1].expr->base.type, False);
			yyval.expr = NewExprTree (NEW, yyvsp[-1].expr, 0); 
			yyval.expr->base.type = yyvsp[-1].expr->base.type;
		    }
break;
case 264:
#line 1375 "gram.y"
{ 
			TypePtr	t = NewTypeArray (typePoly, yyvsp[-5].expr, False);
			ParseCanonType (t, False);
			yyval.expr = NewExprTree (NEW, yyvsp[-1].expr, 0); 
			if (yyvsp[-1].expr)
			    yyvsp[-1].expr->base.type = t;
			yyval.expr->base.type = t;
		    }
break;
case 265:
#line 1384 "gram.y"
{
			TypePtr t = NewTypeHash (typePoly, yyvsp[-5].type);
			ParseCanonType (t, False);
			yyval.expr = NewExprTree (NEW, yyvsp[-1].expr, 0);
			if (yyvsp[-1].expr)
			    yyvsp[-1].expr->base.type = t;
			yyval.expr->base.type = t;
		    }
break;
case 266:
#line 1393 "gram.y"
{
			ParseCanonType (yyvsp[-2].type, False);
			yyval.expr = NewExprTree (UNION, NewExprAtom (yyvsp[0].atom, 0, False), 0); 
			yyval.expr->base.type = yyvsp[-2].type;
		    }
break;
case 267:
#line 1399 "gram.y"
{ 
			ParseCanonType (yyvsp[-4].type, False);
			yyval.expr = NewExprTree (UNION, NewExprAtom (yyvsp[-2].atom, 0, False), yyvsp[0].expr); 
			yyval.expr->base.type = yyvsp[-4].type;
		    }
break;
case 268:
#line 1405 "gram.y"
{ yyval.expr = BuildCall ("History", "fetch", 1, NewExprConst (TEN_NUM, yyvsp[0].value)); }
break;
case 269:
#line 1407 "gram.y"
{ yyval.expr = NewExprTree (DOLLAR, 0, 0); }
break;
case 270:
#line 1409 "gram.y"
{ yyval.expr = yyvsp[-1].expr; }
break;
case 271:
#line 1411 "gram.y"
{ yyval.expr = yyvsp[-1].expr; }
break;
case 272:
#line 1413 "gram.y"
{ yyval.expr = NewExprTree (OS, NewExprTree (STAR, yyvsp[-3].expr, (Expr *) 0), yyvsp[-1].expr); }
break;
case 273:
#line 1415 "gram.y"
{ yyval.expr = NewExprTree(OS, yyvsp[-3].expr, yyvsp[-1].expr); }
break;
case 274:
#line 1417 "gram.y"
{ yyval.expr = NewExprTree (OP, yyvsp[-3].expr, yyvsp[-1].expr); }
break;
case 275:
#line 1419 "gram.y"
{ yyval.expr = NewExprTree(DOT, yyvsp[-2].expr, NewExprAtom (yyvsp[0].atom, 0, False)); }
break;
case 276:
#line 1421 "gram.y"
{ yyval.expr = NewExprTree(ARROW, yyvsp[-2].expr, NewExprAtom (yyvsp[0].atom, 0, False)); }
break;
case 278:
#line 1425 "gram.y"
{ yyval.value = Zero; }
break;
case 298:
#line 1451 "gram.y"
{ yyval.expr = 0; }
break;
case 299:
#line 1453 "gram.y"
{ yyval.expr = 0; }
break;
case 300:
#line 1456 "gram.y"
{ 
			ExprPtr	elts = yyvsp[-3].expr ? ExprRehang (yyvsp[-3].expr, 0) : 0;
			if (yyvsp[-1].bool)
			{
			    ExprPtr i = elts;
			    while (i->tree.right)
				i = i->tree.right;
			    i->tree.right = NewExprTree (COMMA, 
							 NewExprTree (DOTDOTDOT, 0, 0),
							 0);
			}
			yyval.expr = NewExprTree (ARRAY, elts, 0); 
		    }
break;
case 301:
#line 1470 "gram.y"
{
			yyval.expr = NewExprTree (COMP, NewExprTree (COMP, yyvsp[-5].expr, yyvsp[-4].expr), yyvsp[-2].expr);
		    }
break;
case 302:
#line 1475 "gram.y"
{ yyval.expr = yyvsp[0].expr; }
break;
case 304:
#line 1479 "gram.y"
{ yyval.expr = NewExprTree (COMMA, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 306:
#line 1483 "gram.y"
{ 
			SymbolPtr   s;
			s = ParseNewSymbol (publish_private, class_arg, 
					    typePrim[rep_integer], yyvsp[0].atom);
			yyval.expr = NewExprAtom (yyvsp[0].atom, s, False); 
		    }
break;
case 307:
#line 1491 "gram.y"
{
			SymbolPtr   s;
			Atom	    a = AtomId("[]");
			s = ParseNewSymbol (publish_private, class_undef,
					    typePoly, a);
			yyval.expr = NewExprAtom (a, s, False);
		    }
break;
case 308:
#line 1500 "gram.y"
{ yyval.expr = NewExprTree (COMMA, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 309:
#line 1502 "gram.y"
{ yyval.expr = NewExprTree (COMMA, 0, yyvsp[0].expr); }
break;
case 313:
#line 1513 "gram.y"
{ yyval.expr = 0; }
break;
case 314:
#line 1515 "gram.y"
{ yyval.expr = 0; }
break;
case 315:
#line 1518 "gram.y"
{
			ExprPtr elts = yyvsp[-2].expr ? ExprRehang (yyvsp[-2].expr, 0) : 0;
			yyval.expr = NewExprTree (HASH, elts, 0);
		    }
break;
case 316:
#line 1524 "gram.y"
{ yyval.expr = NewExprTree (COMMA, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 317:
#line 1526 "gram.y"
{ yyval.expr = NewExprTree (COMMA, 0, yyvsp[0].expr); }
break;
case 318:
#line 1529 "gram.y"
{ yyval.expr = NewExprTree (DARROW, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 319:
#line 1531 "gram.y"
{ yyval.expr = NewExprTree (DARROW, 0, yyvsp[0].expr); }
break;
case 322:
#line 1540 "gram.y"
{ yyval.expr = NewExprTree (STRUCT, ExprRehang (yyvsp[-2].expr, 0), 0); }
break;
case 323:
#line 1543 "gram.y"
{ yyval.expr = NewExprTree (COMMA, yyvsp[-2].expr, yyvsp[0].expr); }
break;
case 324:
#line 1545 "gram.y"
{ yyval.expr = NewExprTree (COMMA, 0, yyvsp[0].expr); }
break;
case 325:
#line 1548 "gram.y"
{ yyval.expr = NewExprTree (ASSIGN, NewExprAtom (yyvsp[-2].atom, 0, False), yyvsp[0].expr); }
break;
case 326:
#line 1550 "gram.y"
{ yyval.expr = NewExprTree (ASSIGN, NewExprAtom (yyvsp[-2].atom, 0, False), yyvsp[0].expr); }
break;
case 330:
#line 1557 "gram.y"
{ yyval.expr = NewExprTree (ANONINIT, 0, 0); }
break;
#line 4321 "gram.c"
    }
    yyssp -= yym;
    yystate = *yyssp;
    yyvsp -= yym;
    yym = yylhs[yyn];
    if (yystate == 0 && yym == 0)
    {
#if YYDEBUG
        if (yydebug)
            printf("%sdebug: after reduction, shifting from state 0 to\
 state %d\n", YYPREFIX, YYFINAL);
#endif
        yystate = YYFINAL;
        *++yyssp = YYFINAL;
        *++yyvsp = yyval;
        if (yychar < 0)
        {
            if ((yychar = yylex()) < 0) yychar = 0;
#if YYDEBUG
            if (yydebug)
            {
                yys = 0;
                if (yychar <= YYMAXTOKEN) yys = yyname[yychar];
                if (!yys) yys = "illegal-symbol";
                printf("%sdebug: state %d, reading %d (%s)\n",
                        YYPREFIX, YYFINAL, yychar, yys);
            }
#endif
        }
        if (yychar == 0) goto yyaccept;
        goto yyloop;
    }
    if ((yyn = yygindex[yym]) && (yyn += yystate) >= 0 &&
            yyn <= YYTABLESIZE && yycheck[yyn] == yystate)
        yystate = yytable[yyn];
    else
        yystate = yydgoto[yym];
#if YYDEBUG
    if (yydebug)
        printf("%sdebug: after reduction, shifting from state %d \
to state %d\n", YYPREFIX, *yyssp, yystate);
#endif
    if (yyssp >= yyss + yystacksize - 1)
    {
        goto yyoverflow;
    }
    *++yyssp = yystate;
    *++yyvsp = yyval;
    goto yyloop;
yyoverflow:
    yyerror("yacc stack overflow");
yyabort:
    return (1);
yyaccept:
    return (0);
}
